# Copyright (c) OpenMMLab. All rights reserved.
import logging
import re
from typing import Sequence

import numpy as np
import torch
from mmengine.visualization import Visualizer

from mmagic.registry import VISUALIZERS
from mmagic.structures import DataSample
from mmagic.utils import print_colored_log


@VISUALIZERS.register_module()
class ConcatImageVisualizer(Visualizer):
    """Visualize multiple images by concatenation.

    This visualizer will horizontally concatenate images belongs to different
    keys and vertically concatenate images belongs to different frames to
    visualize.

    Image to be visualized can be:
        - torch.Tensor or np.array
        - Image sequences of shape (T, C, H, W)
        - Multi-channel image of shape (1/3, H, W)
        - Single-channel image of shape (C, H, W)

    Args:
        fn_key (str): key used to determine file name for saving image.
            Usually it is the path of some input image. If the value is
            `dir/basename.ext`, the name used for saving will be basename.
        img_keys (str): keys, values of which are images to visualize.
        pixel_range (dict): min and max pixel value used to denormalize images,
            note that only float array or tensor will be denormalized,
            uint8 arrays are assumed to be unnormalized.
        bgr2rgb (bool): whether to convert the image from BGR to RGB.
        name (str): name of visualizer. Default: 'visualizer'.
        *args and \**kwargs: Other arguments are passed to `Visualizer`. # noqa
    """

    def __init__(self,
                 fn_key: str,
                 img_keys: Sequence[str],
                 pixel_range={},
                 bgr2rgb=False,
                 name: str = 'visualizer',
                 *args,
                 **kwargs) -> None:
        super().__init__(name, *args, **kwargs)
        self.fn_key = fn_key
        self.img_keys = img_keys
        self.pixel_range = pixel_range
        self.bgr2rgb = bgr2rgb

    def add_datasample(self, data_sample: DataSample, step=0) -> None:
        """Concatenate image and draw.

        Args:
            input (torch.Tensor): Single input tensor from data_batch.
            data_sample (DataSample): Single data_sample from data_batch.
            output (DataSample): Single prediction output by model.
            step (int): Global step value to record. Default: 0.
        """
        # Note:
        # with LocalVisBackend and default arguments, we have:
        # self.save_dir == runner._log_dir / 'vis_data'

        merged_dict = {
            **data_sample.to_dict(),
        }

        if 'output' in merged_dict.keys():
            merged_dict.update(**merged_dict['output'])

        fn = merged_dict[self.fn_key]
        if isinstance(fn, list):
            fn = fn[0]
        fn = re.split(r' |/|\\', fn)[-1]
        fn = fn.split('.')[0]

        img_list = []
        for k in self.img_keys:
            if k not in merged_dict:
                print_colored_log(
                    f'Key "{k}" not in data_sample or outputs',
                    level=logging.WARN)
                continue

            img = merged_dict[k]

            # PixelData
            if isinstance(img, dict) and ('data' in img):
                img = img['data']

            # Tensor to array
            if isinstance(img, torch.Tensor):
                img = img.detach().cpu().numpy()
                if img.ndim == 3:
                    img = img.transpose(1, 2, 0)
                elif img.ndim == 4:
                    img = img.transpose(0, 2, 3, 1)

            # concat frame vertically
            if img.ndim == 4:
                img = np.concatenate(img, axis=0)

            # gray to 3 channel
            if (img.ndim == 3 and img.shape[2] == 1):
                img = np.concatenate((img, img, img), axis=2)

            # gray to 3 channel
            if img.ndim == 2:
                img = np.stack((img, img, img), axis=2)

            if self.bgr2rgb:
                img = img[..., ::-1]

            if img.dtype != np.uint8:
                # We assume uint8 type are not normalized
                if k in self.pixel_range:
                    min_, max_ = self.pixel_range.get(k)
                    img = ((img - min_) / (max_ - min_)) * 255
                img = img.clip(0, 255).round().astype(np.uint8)

            img_list.append(img)

        max_height = max(img.shape[0] for img in img_list)

        for i, img in enumerate(img_list):
            if img.shape[0] < max_height:
                img_list[i] = np.concatenate([
                    img,
                    np.ones((max_height - img.shape[0], *img.shape[1:]),
                            dtype=img.dtype) * 127
                ],
                                             axis=0)

        img_cat = np.concatenate(img_list, axis=1)

        for vis_backend in self._vis_backends.values():
            vis_backend.add_image(fn, img_cat, step)



# Copyright (c) OpenMMLab. All rights reserved.
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from diffusers import Transformer2DModel
from diffusers.models.attention import Attention, BasicTransformerBlock
from diffusers.models.unet_2d_condition import UNet2DConditionOutput
from diffusers.utils import BaseOutput, is_torch_version


class ViCoCrossAttnProcessor:
    """Processor for implementing attention for the ViCo method."""

    def __call__(self,
                 attn: Attention,
                 hidden_states,
                 encoder_hidden_states=None,
                 attention_mask=None):
        """
        Args:
            attn (Attention): Attention module.
            hidden_states (torch.Tensor): Input hidden states.
            encoder_hidden_states (torch.Tensor): Encoder hidden states.
            attention_mask (torch.Tensor): Attention mask.
        Returns:
            torch.Tensor: Output hidden states.
        """
        batch_size, sequence_length, _ = hidden_states.shape
        attention_mask = attn.prepare_attention_mask(
            attention_mask, sequence_length, batch_size=batch_size)

        query = attn.to_q(hidden_states)

        encoder_hidden_states = encoder_hidden_states \
            if encoder_hidden_states is not None else hidden_states
        key = attn.to_k(encoder_hidden_states)
        value = attn.to_v(encoder_hidden_states)

        query = attn.head_to_batch_dim(query)
        key = attn.head_to_batch_dim(key)
        value = attn.head_to_batch_dim(value)

        attention_probs = attn.get_attention_scores(query, key, attention_mask)
        # new bookkeeping to save the attn probs
        attn.attn_probs = attention_probs

        hidden_states = torch.bmm(attention_probs, value)
        hidden_states = attn.batch_to_head_dim(hidden_states)

        # linear proj
        hidden_states = attn.to_out[0](hidden_states)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        return hidden_states


def replace_cross_attention(unet):
    """Replace Cross Attention processor in UNet."""
    for name, module in unet.named_modules():
        name: str
        if name.endswith('attn2'):
            module.set_processor(ViCoCrossAttnProcessor())


@dataclass
class ViCoTransformer2DModelOutput(BaseOutput):
    """Output for ViCoTransformer2DModel."""
    sample: torch.FloatTensor
    loss_reg: torch.FloatTensor


def otsu(mask_in):
    """Apply otsu for mask.

    Args:
        mask_in (torch.Tensor): Input mask.
    """

    # normalize
    mask_norm = (mask_in - mask_in.min(-1, keepdim=True)[0]) / \
        (mask_in.max(-1, keepdim=True)[0] - mask_in.min(-1, keepdim=True)[0])

    bs = mask_in.shape[0]
    h = mask_in.shape[1]
    mask = []
    for i in range(bs):
        threshold_t = 0.
        max_g = 0.
        for t in range(10):
            mask_i = mask_norm[i]
            low = mask_i[mask_i < t * 0.1]
            high = mask_i[mask_i >= t * 0.1]
            low_num = low.shape[0] / h
            high_num = high.shape[0] / h
            low_mean = low.mean()
            high_mean = high.mean()

            g = low_num * high_num * ((low_mean - high_mean)**2)
            if g > max_g:
                max_g = g
                threshold_t = t * 0.1

        mask_i[mask_i < threshold_t] = 0
        mask_i[mask_i > threshold_t] = 1
        mask.append(mask_i)
    mask_out = torch.stack(mask, dim=0)

    return mask_out


class ViCoTransformer2D(nn.Module):
    """New ViCo-Transformer2D to replace the original Transformer2D model."""

    def __init__(self, org_transformer2d: Transformer2DModel,
                 have_image_cross) -> None:
        """
        Args:
            org_transformer2d (Transformer2DModel): Original
            Transformer2DModel.
            have_image_cross (bool): Flag indicating if the model has
            image_cross_attention modules.
        """
        super().__init__()
        self.transformer_blocks = org_transformer2d.transformer_blocks
        self.is_input_continuous = org_transformer2d.is_input_continuous
        self.norm = org_transformer2d.norm
        self.use_linear_projection = org_transformer2d.use_linear_projection
        self.proj_in = org_transformer2d.proj_in
        self.proj_out = org_transformer2d.proj_out
        self.is_input_vectorized = org_transformer2d.is_input_vectorized
        self.is_input_patches = org_transformer2d.is_input_patches

        num_attention_heads = org_transformer2d.num_attention_heads
        attention_head_dim = org_transformer2d.attention_head_dim
        inner_dim = num_attention_heads * attention_head_dim

        self.have_image_cross = have_image_cross
        if self.have_image_cross:
            image_cross_attention = BasicTransformerBlock(
                inner_dim,
                num_attention_heads,
                attention_head_dim,
                cross_attention_dim=inner_dim)
            self.image_cross_attention = image_cross_attention.to(
                org_transformer2d.device, dtype=org_transformer2d.dtype)

    def forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        timestep: Optional[torch.LongTensor] = None,
        placeholder_position: list = None,
        class_labels: Optional[torch.LongTensor] = None,
        cross_attention_kwargs: Dict[str, Any] = None,
        attention_mask: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        return_dict: bool = True,
    ):
        if attention_mask is not None and attention_mask.ndim == 2:
            # assume that mask is expressed as:
            #   (1 = keep,      0 = discard)
            # convert mask into a bias that can be added to attention scores:
            #       (keep = +0,     discard = -10000.0)
            attention_mask = (
                1 - attention_mask.to(hidden_states.dtype)) * -10000.0
            attention_mask = attention_mask.unsqueeze(1)

        if encoder_attention_mask is not None and (encoder_attention_mask.ndim
                                                   == 2):
            encoder_attention_mask = (
                1 - encoder_attention_mask.to(hidden_states.dtype)) * -10000.0
            encoder_attention_mask = encoder_attention_mask.unsqueeze(1)

        # 1. Input
        if self.is_input_continuous:
            batch, _, height, width = hidden_states.shape
            residual = hidden_states

            hidden_states = self.norm(hidden_states)
            if not self.use_linear_projection:
                hidden_states = self.proj_in(hidden_states)
                inner_dim = hidden_states.shape[1]
                hidden_states = hidden_states.permute(0, 2, 3, 1).reshape(
                    batch, height * width, inner_dim)
            else:
                inner_dim = hidden_states.shape[1]
                hidden_states = hidden_states.permute(0, 2, 3, 1).reshape(
                    batch, height * width, inner_dim)
                hidden_states = self.proj_in(hidden_states)
        elif self.is_input_vectorized:
            hidden_states = self.latent_image_embedding(hidden_states)
        elif self.is_input_patches:
            hidden_states = self.pos_embed(hidden_states)

        # 2. Blocks
        for block in self.transformer_blocks:
            hidden_states = block(
                hidden_states,
                attention_mask=attention_mask,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                timestep=timestep,
                cross_attention_kwargs=cross_attention_kwargs,
                class_labels=class_labels,
            )
            attention_probs = block.attn2.attn_probs[batch // 2:batch, ...]

            loss_reg = None
            if self.have_image_cross:
                # 2.5. image cross attention
                ph_idx, eot_idx = placeholder_position[
                    0], placeholder_position[1]
                attn = attention_probs.transpose(1, 2)
                attn_ph = attn[ph_idx].squeeze(1)  # bs, n_patch
                attn_eot = attn[eot_idx].squeeze(1).detach()

                # ########################
                # attention reg
                if self.image_cross_attention.training:
                    loss_reg = F.mse_loss(
                        attn_ph / attn_ph.max(-1, keepdim=True)[0],
                        attn_eot / attn_eot.max(-1, keepdim=True)[0])
                # ########################

                mask = attn_ph.detach()
                mask = otsu(mask)
                if mask.dim() == 2:
                    mask = mask.unsqueeze(1)

                hidden_states, image_reference = hidden_states[:batch // 2], \
                    hidden_states[batch // 2:]
                hidden_states = self.image_cross_attention(
                    hidden_states,
                    attention_mask=attention_mask,
                    encoder_hidden_states=image_reference,
                    encoder_attention_mask=mask,
                    timestep=timestep,
                    cross_attention_kwargs=cross_attention_kwargs,
                    class_labels=class_labels,
                )
                hidden_states = torch.cat([hidden_states, image_reference],
                                          dim=0)

        # 3. Output
        if self.is_input_continuous:
            if not self.use_linear_projection:
                hidden_states = hidden_states.reshape(batch, height, width,
                                                      inner_dim).permute(
                                                          0, 3, 1,
                                                          2).contiguous()
                hidden_states = self.proj_out(hidden_states)
            else:
                hidden_states = self.proj_out(hidden_states)
                hidden_states = hidden_states.reshape(batch, height, width,
                                                      inner_dim).permute(
                                                          0, 3, 1,
                                                          2).contiguous()

            output = hidden_states + residual
        elif self.is_input_vectorized:
            hidden_states = self.norm_out(hidden_states)
            logits = self.out(hidden_states)
            # (batch, self.num_vector_embeds - 1, self.num_latent_pixels)
            logits = logits.permute(0, 2, 1)

            # log(p(x_0))
            output = F.log_softmax(logits.double(), dim=1).float()
        elif self.is_input_patches:
            # TODO: cleanup!
            conditioning = self.transformer_blocks[0].norm1.emb(
                timestep, class_labels, hidden_dtype=hidden_states.dtype)
            shift, scale = self.proj_out_1(F.silu(conditioning)).chunk(
                2, dim=1)
            hidden_states = self.norm_out(hidden_states) * (
                1 + scale[:, None]) + shift[:, None]
            hidden_states = self.proj_out_2(hidden_states)

            # unpatchify
            height = width = int(hidden_states.shape[1]**0.5)
            hidden_states = hidden_states.reshape(
                shape=(-1, height, width, self.patch_size, self.patch_size,
                       self.out_channels))
            hidden_states = torch.einsum('nhwpqc->nchpwq', hidden_states)
            output = hidden_states.reshape(
                shape=(-1, self.out_channels, height * self.patch_size,
                       width * self.patch_size))

        if not return_dict:
            return (output, loss_reg)

        return ViCoTransformer2DModelOutput(sample=output, loss_reg=loss_reg)


def replace_transformer2d(module: nn.Module,
                          have_image_cross: Dict[str, List[bool]]):
    """Replace the the Transformer2DModel in UNet.

    Args:
        module (nn.Module): Parent module of Transformer2D.
        have_image_cross (List): List of flag indicating which
        transformer2D modules have image_cross_attention modules.
    """
    down_transformer2d_modules = [(k.rsplit('.', 1), v)
                                  for k, v in module.named_modules()
                                  if isinstance(v, Transformer2DModel)]
    for i, ((parent, k), v) in enumerate(down_transformer2d_modules):
        parent = module.get_submodule(parent)
        setattr(parent, k, ViCoTransformer2D(v, have_image_cross[i]))


class ViCoBlockWrapper(nn.Module):
    """Wrapper for ViCo blocks."""

    def apply_to(self, org_module):
        self.org_module = org_module
        self.org_module.forward = self.forward


class ViCoCrossAttnDownBlock2D(ViCoBlockWrapper):

    def forward(
        self,
        hidden_states: torch.FloatTensor,
        temb: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        placeholder_position: torch.Tensor = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
    ):
        """
        Args:
            hidden_states (torch.FloatTensor): Hidden states.
            temb (Optional[torch.FloatTensor]): Time embedding.
            encoder_hidden_states (Optional[torch.FloatTensor]): Encoder
                hidden states.
            placeholder_position (torch.Tensor): Placeholder position.
            attention_mask (Optional[torch.FloatTensor]): Attention mask.
            cross_attention_kwargs (Optional[Dict[str, Any]]): Cross attention
                keyword arguments.
            encoder_attention_mask (Optional[torch.FloatTensor]): Encoder
                attention mask.
        Returns:
            torch.FloatTensor: Output hidden states.
            Tuple[torch.FloatTensor]: Output hidden states of each block.
            torch.FloatTensor: Attention regularization loss.
        """
        output_states = ()
        loss_reg_all = 0.0

        for resnet, attn in zip(self.org_module.resnets,
                                self.org_module.attentions):
            attn: ViCoTransformer2D
            if self.org_module.training and (
                    self.org_module.gradient_checkpointing):

                def create_custom_forward(module, return_dict=None):

                    def custom_forward(*inputs):
                        if return_dict is not None:
                            return module(*inputs, return_dict=return_dict)
                        else:
                            return module(*inputs)

                    return custom_forward

                ckpt_kwargs: Dict[str, Any] = {
                    'use_reentrant': False
                } if is_torch_version('>=', '1.11.0') else {}
                hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(resnet),
                    hidden_states,
                    temb,
                    **ckpt_kwargs,
                )
                hidden_states, loss_reg = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(attn, return_dict=False),
                    hidden_states,
                    encoder_hidden_states,
                    None,  # timestep
                    placeholder_position,
                    None,  # class_labels
                    cross_attention_kwargs,
                    attention_mask,
                    encoder_attention_mask,
                    **ckpt_kwargs,
                )
            else:
                hidden_states = resnet(hidden_states, temb)
                hidden_states, loss_reg = attn(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    placeholder_position=placeholder_position,
                    cross_attention_kwargs=cross_attention_kwargs,
                    attention_mask=attention_mask,
                    encoder_attention_mask=encoder_attention_mask,
                    return_dict=False,
                )

            output_states = output_states + (hidden_states, )
            if loss_reg is not None:
                loss_reg_all += loss_reg

        if self.org_module.downsamplers is not None:
            for downsampler in self.org_module.downsamplers:
                hidden_states = downsampler(hidden_states)

            output_states = output_states + (hidden_states, )

        return hidden_states, output_states, loss_reg_all


class ViCoUNetMidBlock2DCrossAttn(ViCoBlockWrapper):

    def forward(
        self,
        hidden_states: torch.FloatTensor,
        temb: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        placeholder_position: torch.Tensor = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
    ) -> torch.FloatTensor:
        """
        Args:
            hidden_states (torch.FloatTensor): Hidden states.
            temb (Optional[torch.FloatTensor]): Time embedding.
            encoder_hidden_states (Optional[torch.FloatTensor]): Encoder
                hidden states.
            placeholder_position (torch.Tensor): Placeholder position.
            attention_mask (Optional[torch.FloatTensor]): Attention mask.
            cross_attention_kwargs (Optional[Dict[str, Any]]): Cross attention
                keyword arguments.
            encoder_attention_mask (Optional[torch.FloatTensor]): Encoder
                attention mask.
        Returns:
            torch.FloatTensor: Output hidden states.
            torch.FloatTensor: Attention regularization loss.
        """
        loss_reg_all = 0.0
        hidden_states = self.org_module.resnets[0](hidden_states, temb)
        for attn, resnet in zip(self.org_module.attentions,
                                self.org_module.resnets[1:]):
            hidden_states, loss_reg = attn(
                hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                placeholder_position=placeholder_position,
                cross_attention_kwargs=cross_attention_kwargs,
                attention_mask=attention_mask,
                encoder_attention_mask=encoder_attention_mask,
                return_dict=False,
            )
            hidden_states = resnet(hidden_states, temb)
            if loss_reg is not None:
                loss_reg_all += loss_reg

        return hidden_states, loss_reg_all


class ViCoCrossAttnUpBlock2D(ViCoBlockWrapper):

    def forward(
        self,
        hidden_states: torch.FloatTensor,
        res_hidden_states_tuple: Tuple[torch.FloatTensor, ...],
        temb: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        placeholder_position: torch.Tensor = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        upsample_size: Optional[int] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
    ):
        """Performs the forward pass through the ViCoCrossAttnUpBlock2D module.

        Args:
            hidden_states (torch.FloatTensor): Input hidden states.
            res_hidden_states_tuple (Tuple[torch.FloatTensor, ...]):
                Tuple of residual hidden states.
            temb (Optional[torch.FloatTensor], optional):
                Temporal embeddings. Defaults to None.
            encoder_hidden_states (Optional[torch.FloatTensor], optional):
                Encoder hidden states. Defaults to None.
            placeholder_position (torch.Tensor, optional):
                Placeholder positions. Defaults to None.
            cross_attention_kwargs (Optional[Dict[str, Any]], optional):
                Keyword arguments for cross-attention. Defaults to None.
            upsample_size (Optional[int], optional): Upsample size.
            attention_mask (Optional[torch.FloatTensor], optional):
              Attention mask.
            encoder_attention_mask (Optional[torch.FloatTensor], optional):
                Encoder attention mask.

        Returns:
            Tuple[torch.FloatTensor, torch.FloatTensor]:
                A tuple containing the output hidden states and
                the total regularization loss.
        """
        loss_reg_all = 0.0
        for resnet, attn in zip(self.org_module.resnets,
                                self.org_module.attentions):
            attn: ViCoTransformer2D

            # pop res hidden states
            res_hidden_states = res_hidden_states_tuple[-1]
            res_hidden_states_tuple = res_hidden_states_tuple[:-1]
            hidden_states = torch.cat([hidden_states, res_hidden_states],
                                      dim=1)

            if self.org_module.training and (
                    self.org_module.gradient_checkpointing):

                def create_custom_forward(module, return_dict=None):

                    def custom_forward(*inputs):
                        if return_dict is not None:
                            return module(*inputs, return_dict=return_dict)
                        else:
                            return module(*inputs)

                    return custom_forward

                ckpt_kwargs: Dict[str, Any] = {
                    'use_reentrant': False
                } if is_torch_version('>=', '1.11.0') else {}
                hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(resnet),
                    hidden_states,
                    temb,
                    **ckpt_kwargs,
                )
                hidden_states, loss_reg = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(attn, return_dict=False),
                    hidden_states,
                    encoder_hidden_states,
                    None,  # timestep
                    placeholder_position,
                    None,  # class_labels
                    cross_attention_kwargs,
                    attention_mask,
                    encoder_attention_mask,
                    **ckpt_kwargs,
                )
            else:
                hidden_states = resnet(hidden_states, temb)
                hidden_states, loss_reg = attn(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    placeholder_position=placeholder_position,
                    cross_attention_kwargs=cross_attention_kwargs,
                    attention_mask=attention_mask,
                    encoder_attention_mask=encoder_attention_mask,
                    return_dict=False,
                )
            if loss_reg is not None:
                loss_reg_all += loss_reg

        if self.org_module.upsamplers is not None:
            for upsampler in self.org_module.upsamplers:
                hidden_states = upsampler(hidden_states, upsample_size)

        return hidden_states, loss_reg_all


class ViCoUNet2DConditionOutput(BaseOutput):
    """Output for ViCoUNet2DConditionModel."""
    sample: torch.FloatTensor
    loss_reg: torch.FloatTensor


class ViCoUNet2DConditionModel(ViCoBlockWrapper):
    """UNet2DConditionModel for ViCo Method."""

    def forward(
        self,
        sample: torch.FloatTensor,
        timestep: Union[torch.Tensor, float, int],
        encoder_hidden_states: torch.Tensor,
        placeholder_position: torch.Tensor,
        class_labels: Optional[torch.Tensor] = None,
        timestep_cond: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        added_cond_kwargs: Optional[Dict[str, torch.Tensor]] = None,
        down_block_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
        mid_block_additional_residual: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        return_dict: bool = True,
    ) -> Union[UNet2DConditionOutput, Tuple]:
        """Performs the forward pass through the ViCoBlock2D module.

        Args:
            sample (torch.FloatTensor): Input sample.
            timestep (Union[torch.Tensor, float, int]): Timestep value.
            encoder_hidden_states (torch.Tensor): Encoder hidden states.
            placeholder_position (torch.Tensor): Placeholder positions.
            class_labels (Optional[torch.Tensor], optional): Class labels.
                Defaults to None.
            timestep_cond (Optional[torch.Tensor], optional):
                Timestep condition. Defaults to None.
            attention_mask (Optional[torch.Tensor], optional):
                Attention mask. Defaults to None.
            cross_attention_kwargs (Optional[Dict[str, Any]], optional):
                Keyword arguments for cross-attention. Defaults to None.
            added_cond_kwargs (Optional[Dict[str, torch.Tensor]], optional):
                Additional condition arguments. Defaults to None.
            down_block_additional_residuals
                (Optional[Tuple[torch.Tensor]], optional):
                Additional residuals for down-blocks. Defaults to None.
            mid_block_additional_residual (Optional[torch.Tensor], optional):
                Additional residual for mid-block. Defaults to None.
            encoder_attention_mask (Optional[torch.Tensor], optional):
                Encoder attention mask. Defaults to None.
            return_dict (bool, optional):
                Whether to return a dictionary or a tuple.

        Returns:
            Union[UNet2DConditionOutput, Tuple]:
                The output of the forward pass, which can be either
                a UNet2DConditionOutput object or a tuple of tensors.
        """
        default_overall_up_factor = 2**self.org_module.num_upsamplers
        forward_upsample_size = False
        upsample_size = None

        if any(s % default_overall_up_factor != 0 for s in sample.shape[-2:]):
            forward_upsample_size = True

        if attention_mask is not None:
            # assume that mask is expressed as:
            #   (1 = keep,      0 = discard)
            # convert mask into a bias that can be added to attention scores:
            #       (keep = +0,     discard = -10000.0)
            attention_mask = (1 - attention_mask.to(sample.dtype)) * -10000.0
            attention_mask = attention_mask.unsqueeze(1)

        if encoder_attention_mask is not None:
            encoder_attention_mask = (
                1 - encoder_attention_mask.to(sample.dtype)) * -10000.0
            encoder_attention_mask = encoder_attention_mask.unsqueeze(1)

        # 0. center input if necessary
        if self.org_module.config.center_input_sample:
            sample = 2 * sample - 1.0

        # 1. time
        timesteps = timestep
        if not torch.is_tensor(timesteps):
            is_mps = sample.device.type == 'mps'
            if isinstance(timestep, float):
                dtype = torch.float32 if is_mps else torch.float64
            else:
                dtype = torch.int32 if is_mps else torch.int64
            timesteps = torch.tensor([timesteps],
                                     dtype=dtype,
                                     device=sample.device)
        elif len(timesteps.shape) == 0:
            timesteps = timesteps[None].to(sample.device)

        t_emb = self.org_module.time_proj(timesteps)

        t_emb = t_emb.to(dtype=sample.dtype)

        emb = self.org_module.time_embedding(t_emb, timestep_cond)

        if self.org_module.class_embedding is not None:
            if class_labels is None:
                raise ValueError('class_labels should be provided \
                    when num_class_embeds > 0')

            if self.org_module.config.class_embed_type == 'timestep':
                class_labels = self.org_module.time_proj(class_labels)
                class_labels = class_labels.to(dtype=sample.dtype)

            class_emb = self.org_module.class_embedding(class_labels).to(
                dtype=sample.dtype)

            if self.org_module.config.class_embeddings_concat:
                emb = torch.cat([emb, class_emb], dim=-1)
            else:
                emb = emb + class_emb

        if self.org_module.config.addition_embed_type == 'text':
            aug_emb = self.org_module.add_embedding(encoder_hidden_states)
            emb = emb + aug_emb
        elif self.org_module.config.addition_embed_type == 'text_image':
            # Kadinsky 2.1 - style
            if 'image_embeds' not in added_cond_kwargs:
                raise ValueError(
                    f"{self.org_module.__class__} has the config param \
                    `addition_embed_type` set to 'text_image' which \
                    requires the keyword argument `image_embeds` \
                    to be passed in `added_cond_kwargs`")

            image_embs = added_cond_kwargs.get('image_embeds')
            text_embs = added_cond_kwargs.get('text_embeds',
                                              encoder_hidden_states)

            aug_emb = self.org_module.add_embedding(text_embs, image_embs)
            emb = emb + aug_emb

        if self.org_module.time_embed_act is not None:
            emb = self.org_module.time_embed_act(emb)

        if self.org_module.encoder_hid_proj is not None and (
                self.org_module.config.encoder_hid_dim_type == 'text_proj'):
            encoder_hidden_states = self.org_module.encoder_hid_proj(
                encoder_hidden_states)
        elif self.org_module.encoder_hid_proj is not None and (
                self.org_module.config.encoder_hid_dim_type
                == 'text_image_proj'):
            # Kadinsky 2.1 - style
            if 'image_embeds' not in added_cond_kwargs:
                raise ValueError(
                    f"{self.org_module.__class__} has the config param \
                        `encoder_hid_dim_type` set to 'text_image_proj' which \
                        requires the keyword argument `image_embeds` to be \
                        passed in  `added_conditions`")

            image_embeds = added_cond_kwargs.get('image_embeds')
            encoder_hidden_states = self.org_module.encoder_hid_proj(
                encoder_hidden_states, image_embeds)

        # 2. pre-process
        sample = self.org_module.conv_in(sample)

        loss_reg_all = 0.0 if self.training else None

        # 3. down
        down_block_res_samples = (sample, )
        for downsample_block in self.org_module.down_blocks:
            if hasattr(downsample_block, 'has_cross_attention'
                       ) and downsample_block.has_cross_attention:
                sample, res_samples, loss_reg = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    encoder_hidden_states=encoder_hidden_states,
                    placeholder_position=placeholder_position,
                    attention_mask=attention_mask,
                    cross_attention_kwargs=cross_attention_kwargs,
                    encoder_attention_mask=encoder_attention_mask,
                )
                if self.training:
                    loss_reg_all += loss_reg
            else:
                sample, res_samples = downsample_block(
                    hidden_states=sample, temb=emb)

            down_block_res_samples += res_samples

        if down_block_additional_residuals is not None:
            new_down_block_res_samples = ()

            for down_block_res_sample, down_block_additional_residual in zip(
                    down_block_res_samples, down_block_additional_residuals):
                down_block_res_sample = down_block_res_sample + \
                    down_block_additional_residual
                new_down_block_res_samples = new_down_block_res_samples + (
                    down_block_res_sample, )

            down_block_res_samples = new_down_block_res_samples

        # 4. mid
        if self.org_module.mid_block is not None:
            sample, loss_reg = self.org_module.mid_block(
                sample,
                emb,
                encoder_hidden_states=encoder_hidden_states,
                attention_mask=attention_mask,
                cross_attention_kwargs=cross_attention_kwargs,
                encoder_attention_mask=encoder_attention_mask,
            )
            if self.training:
                loss_reg_all += loss_reg
        if mid_block_additional_residual is not None:
            sample = sample + mid_block_additional_residual

        # 5. up
        for i, upsample_block in enumerate(self.org_module.up_blocks):
            is_final_block = i == len(self.org_module.up_blocks) - 1

            res_samples = down_block_res_samples[-len(upsample_block.resnets):]
            down_block_res_samples = down_block_res_samples[:-len(
                upsample_block.resnets)]

            # if we have not reached the final block and need to forward the
            # upsample size, we do it here
            if not is_final_block and forward_upsample_size:
                upsample_size = down_block_res_samples[-1].shape[2:]

            if hasattr(upsample_block, 'has_cross_attention'
                       ) and upsample_block.has_cross_attention:
                sample, loss_reg = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    encoder_hidden_states=encoder_hidden_states,
                    placeholder_position=placeholder_position,
                    cross_attention_kwargs=cross_attention_kwargs,
                    upsample_size=upsample_size,
                    attention_mask=attention_mask,
                    encoder_attention_mask=encoder_attention_mask,
                )
                if self.training:
                    loss_reg_all += loss_reg
            else:
                sample = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    upsample_size=upsample_size)

        # 6. post-process
        if self.org_module.conv_norm_out:
            sample = self.org_module.conv_norm_out(sample)
            sample = self.org_module.conv_act(sample)
        sample = self.org_module.conv_out(sample)

        if not return_dict:
            return (sample, loss_reg_all)

        return ViCoUNet2DConditionOutput(sample=sample, loss_reg=loss_reg_all)


def set_vico_modules(unet, image_cross_layers):
    """Set all modules for ViCo method after the UNet initialized normally.

    Args:
        unet (nn.Module): UNet model.
        image_cross_layers (List): List of flag indicating which
        transformer2D modules have image_cross_attention modules.
    """
    # replace transformer2d blocks
    replace_transformer2d(unet, image_cross_layers)

    # replace cross attention layer
    replace_cross_attention(unet)

    # replace forward
    for _, layer in unet.named_modules():
        if layer.__class__.__name__ == 'UNet2DConditionModel':
            vico_unet = ViCoUNet2DConditionModel()
            vico_unet.apply_to(unet)
        elif layer.__class__.__name__ == 'CrossAttnDownBlock2D':
            vico_down_block = ViCoCrossAttnDownBlock2D()
            vico_down_block.apply_to(layer)
        elif layer.__class__.__name__ == 'UNetMidBlock2DCrossAttn':
            vico_mid_block = ViCoUNetMidBlock2DCrossAttn()
            vico_mid_block.apply_to(layer)
        elif layer.__class__.__name__ == 'CrossAttnUpBlock2D':
            vico_up_block = ViCoCrossAttnUpBlock2D()
            vico_up_block.apply_to(layer)



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine.runner import LogProcessor as BaseLogProcessor

from mmagic.registry import LOG_PROCESSORS


@LOG_PROCESSORS.register_module()  # type: ignore
class LogProcessor(BaseLogProcessor):
    """LogProcessor inherits from :class:`mmengine.runner.LogProcessor` and
    overwrites :meth:`self.get_log_after_iter`.

    This log processor should be used along with
    :class:`mmagic.engine.runner.MultiValLoop` and
    :class:`mmagic.engine.runner.MultiTestLoop`.
    """

    def _get_dataloader_size(self, runner, mode) -> int:
        """Get dataloader size of current loop. In `MultiValLoop` and
        `MultiTestLoop`, we use `total_length` instead of `len(dataloader)` to
        denote the total number of iterations.

        Args:
            runner (Runner): The runner of the training/validation/testing
            mode (str): Current mode of runner.

        Returns:
            int: The dataloader size of current loop.
        """
        if hasattr(self._get_cur_loop(runner, mode), 'total_length'):
            return self._get_cur_loop(runner, mode).total_length
        else:
            return super()._get_dataloader_size(runner, mode)



_base_ = ['./singan_fish.py']

# MODEL
# NOTE: add by user, e.g.:
# test_pkl_data = './work_dirs/singan_pkl/singan_bohemian_20210406_175439-f964ee38.pkl'  # noqa
test_pkl_data = None
model = dict(test_pkl_data=test_pkl_data)

# HOOKS
custom_hooks = [
    dict(
        type='PickleDataHook',
        output_dir='pickle',
        interval=-1,
        after_run=True,
        data_name_list=['noise_weights', 'fixed_noises', 'curr_stage']),
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        vis_kwargs_list=dict(type='SinGAN', name='bohemian'))
]

# DATA
min_size = 25
max_size = 500
data_root = './data/singan/bohemian.png'
train_dataloader = dict(
    dataset=dict(data_root=data_root, min_size=min_size, max_size=max_size))



# Copyright (c) OpenMMLab. All rights reserved.
"""Augmentation on alpha matte."""
# Not used in current algorithms

import random

import cv2
import numpy as np
from mmcv.transforms import BaseTransform
from mmengine.utils import is_list_of, is_tuple_of

from mmagic.registry import TRANSFORMS
from mmagic.utils import random_choose_unknown


@TRANSFORMS.register_module()
class GenerateSeg(BaseTransform):
    """Generate segmentation mask from alpha matte.

    Args:
        kernel_size (int, optional): Kernel size for both erosion and
            dilation. The kernel will have the same height and width.
            Defaults to 5.
        erode_iter_range (tuple, optional): Iteration of erosion.
            Defaults to (10, 20).
        dilate_iter_range (tuple, optional): Iteration of dilation.
            Defaults to (15, 30).
        num_holes_range (tuple, optional): Range of number of holes to
            randomly select from. Defaults to (0, 3).
        hole_sizes (list, optional): List of (h, w) to be selected as the
            size of the rectangle hole.
            Defaults to [(15, 15), (25, 25), (35, 35), (45, 45)].
        blur_ksizes (list, optional): List of (h, w) to be selected as the
            kernel_size of the gaussian blur.
            Defaults to [(21, 21), (31, 31), (41, 41)].
    """

    def __init__(self,
                 kernel_size=5,
                 erode_iter_range=(10, 20),
                 dilate_iter_range=(15, 30),
                 num_holes_range=(0, 3),
                 hole_sizes=[(15, 15), (25, 25), (35, 35), (45, 45)],
                 blur_ksizes=[(21, 21), (31, 31), (41, 41)]):
        self.kernel_size = kernel_size
        self.erode_iter_range = erode_iter_range
        self.dilate_iter_range = dilate_iter_range
        self.num_holes_range = num_holes_range
        self.hole_sizes = hole_sizes
        self.blur_ksizes = blur_ksizes

    @staticmethod
    def _crop_hole(img, start_point, hole_size):
        """Create a all-zero rectangle hole in the image.

        Args:
            img (np.ndarray): Source image.
            start_point (tuple[int]): The top-left point of the rectangle.
            hole_size (tuple[int]): The height and width of the rectangle hole.

        Return:
            np.ndarray: The cropped image.
        """
        top, left = start_point
        bottom = top + hole_size[0]
        right = left + hole_size[1]
        height, weight = img.shape[:2]
        if top < 0 or bottom > height or left < 0 or right > weight:
            raise ValueError(f'crop area {(left, top, right, bottom)} exceeds '
                             f'image size {(height, weight)}')
        img[top:bottom, left:right] = 0
        return img

    def transform(self, results: dict) -> dict:
        """Transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """
        alpha = results['alpha']
        trimap = results['trimap']

        # generate segmentation mask
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,
                                           (self.kernel_size,
                                            self.kernel_size))
        seg = (alpha > 0.5).astype(np.float32)
        seg = cv2.erode(
            seg, kernel, iterations=np.random.randint(*self.erode_iter_range))
        seg = cv2.dilate(
            seg, kernel, iterations=np.random.randint(*self.dilate_iter_range))

        # generate some holes in segmentation mask
        num_holes = np.random.randint(*self.num_holes_range)
        for _ in range(num_holes):
            hole_size = random.choice(self.hole_sizes)
            unknown = trimap == 128
            start_point = random_choose_unknown(unknown, hole_size)
            seg = self._crop_hole(seg, start_point, hole_size)
            trimap = self._crop_hole(trimap, start_point, hole_size)

        # perform gaussian blur to segmentation mask
        seg = cv2.GaussianBlur(seg, random.choice(self.blur_ksizes), 0)

        results['seg'] = seg.astype(np.uint8)
        results['num_holes'] = num_holes
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (
            f'(kernel_size={self.kernel_size}, '
            f'erode_iter_range={self.erode_iter_range}, '
            f'dilate_iter_range={self.dilate_iter_range}, '
            f'num_holes_range={self.num_holes_range}, '
            f'hole_sizes={self.hole_sizes}, blur_ksizes={self.blur_ksizes}')
        return repr_str


@TRANSFORMS.register_module()
class GenerateSoftSeg(BaseTransform):
    """Generate soft segmentation mask from input segmentation mask.

    Required key is "seg", added key is "soft_seg".

    Args:
        fg_thr (float, optional): Threshold of the foreground in the normalized
            input segmentation mask. Defaults to 0.2.
        border_width (int, optional): Width of border to be padded to the
            bottom of the mask. Defaults to 25.
        erode_ksize (int, optional): Fixed kernel size of the erosion.
            Defaults to 5.
        dilate_ksize (int, optional): Fixed kernel size of the dilation.
            Defaults to 5.
        erode_iter_range (tuple, optional): Iteration of erosion.
            Defaults to (10, 20).
        dilate_iter_range (tuple, optional): Iteration of dilation.
            Defaults to (3, 7).
        blur_ksizes (list, optional): List of (h, w) to be selected as the
            kernel_size of the gaussian blur.
            Defaults to [(21, 21), (31, 31), (41, 41)].
    """

    def __init__(self,
                 fg_thr=0.2,
                 border_width=25,
                 erode_ksize=3,
                 dilate_ksize=5,
                 erode_iter_range=(10, 20),
                 dilate_iter_range=(3, 7),
                 blur_ksizes=[(21, 21), (31, 31), (41, 41)]):
        if not isinstance(fg_thr, float):
            raise TypeError(f'fg_thr must be a float, but got {type(fg_thr)}')
        if not isinstance(border_width, int):
            raise TypeError(
                f'border_width must be an int, but got {type(border_width)}')
        if not isinstance(erode_ksize, int):
            raise TypeError(
                f'erode_ksize must be an int, but got {type(erode_ksize)}')
        if not isinstance(dilate_ksize, int):
            raise TypeError(
                f'dilate_ksize must be an int, but got {type(dilate_ksize)}')
        if (not is_tuple_of(erode_iter_range, int)
                or len(erode_iter_range) != 2):
            raise TypeError('erode_iter_range must be a tuple of 2 int, '
                            f'but got {erode_iter_range}')
        if (not is_tuple_of(dilate_iter_range, int)
                or len(dilate_iter_range) != 2):
            raise TypeError('dilate_iter_range must be a tuple of 2 int, '
                            f'but got {dilate_iter_range}')
        if not is_list_of(blur_ksizes, tuple):
            raise TypeError(
                f'blur_ksizes must be a list of tuple, but got {blur_ksizes}')

        self.fg_thr = fg_thr
        self.border_width = border_width
        self.erode_ksize = erode_ksize
        self.dilate_ksize = dilate_ksize
        self.erode_iter_range = erode_iter_range
        self.dilate_iter_range = dilate_iter_range
        self.blur_ksizes = blur_ksizes

    def transform(self, results: dict) -> dict:
        """Transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """
        seg = results['seg'].astype(np.float32) / 255
        height, _ = seg.shape[:2]
        seg[seg > self.fg_thr] = 1

        # to align with the original repo, pad the bottom of the mask
        seg = cv2.copyMakeBorder(seg, 0, self.border_width, 0, 0,
                                 cv2.BORDER_REPLICATE)

        # erode/dilate segmentation mask
        erode_kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,
                                                 (self.erode_ksize,
                                                  self.erode_ksize))
        dilate_kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,
                                                  (self.dilate_ksize,
                                                   self.dilate_ksize))
        seg = cv2.erode(
            seg,
            erode_kernel,
            iterations=np.random.randint(*self.erode_iter_range))
        seg = cv2.dilate(
            seg,
            dilate_kernel,
            iterations=np.random.randint(*self.dilate_iter_range))

        # perform gaussian blur to segmentation mask
        seg = cv2.GaussianBlur(seg, random.choice(self.blur_ksizes), 0)

        # remove the padded rows
        seg = (seg * 255).astype(np.uint8)
        seg = np.delete(seg, range(height, height + self.border_width), 0)

        results['soft_seg'] = seg
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (f'(fg_thr={self.fg_thr}, '
                     f'border_width={self.border_width}, '
                     f'erode_ksize={self.erode_ksize}, '
                     f'dilate_ksize={self.dilate_ksize}, '
                     f'erode_iter_range={self.erode_iter_range}, '
                     f'dilate_iter_range={self.dilate_iter_range}, '
                     f'blur_ksizes={self.blur_ksizes})')
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
from .multi_optimizer_constructor import MultiOptimWrapperConstructor
from .pggan_optimizer_constructor import PGGANOptimWrapperConstructor
from .singan_optimizer_constructor import SinGANOptimWrapperConstructor

__all__ = [
    'MultiOptimWrapperConstructor',
    'PGGANOptimWrapperConstructor',
    'SinGANOptimWrapperConstructor',
]



# Copyright (c) OpenMMLab. All rights reserved.
import math
from logging import WARNING
from typing import List, Optional, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from mmengine import print_log
from mmengine.model import ImgDataPreprocessor
from mmengine.utils import is_seq_of
from torch import Tensor

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import SampleList

CastData = Union[tuple, dict, DataSample, Tensor, list]


@MODELS.register_module()
class DataPreprocessor(ImgDataPreprocessor):
    """Image pre-processor for generative models. This class provide
    normalization and bgr to rgb conversion for image tensor inputs. The input
    of this classes should be dict which keys are `inputs` and `data_samples`.

    Besides to process tensor `inputs`, this class support dict as `inputs`.
    - If the value is `Tensor` and the corresponding key is not contained in
    :attr:`_NON_IMAGE_KEYS`, it will be processed as image tensor.
    - If the value is `Tensor` and the corresponding key belongs to
    :attr:`_NON_IMAGE_KEYS`, it will not remains unchanged.
    - If value is string or integer, it will not remains unchanged.

    Args:
        mean (Sequence[float or int], float or int, optional): The pixel mean
            of image channels. Noted that normalization operation is performed
            *after channel order conversion*. If it is not specified, images
            will not be normalized. Defaults None.
        std (Sequence[float or int], float or int, optional): The pixel
            standard deviation of image channels. Noted that normalization
            operation is performed *after channel order conversion*. If it is
            not specified, images will not be normalized. Defaults None.
        pad_size_divisor (int): The size of padded image should be
            divisible by ``pad_size_divisor``. Defaults to 1.
        pad_value (float or int): The padded pixel value. Defaults to 0.
        pad_mode (str): Padding mode for ``torch.nn.functional.pad``.
            Defaults to 'constant'.
        non_image_keys (List[str] or str): Keys for fields that not need to be
            processed (padding, channel conversion and normalization) as
            images. If not passed, the keys in :attr:`_NON_IMAGE_KEYS` will be
            used. This argument will only work when `inputs` is dict or list
            of dict. Defaults to None.
        non_concatenate_keys (List[str] or str): Keys for fields that not need
            to be concatenated. If not passed, the keys in
            :attr:`_NON_CONCATENATE_KEYS` will be used. This argument will only
            work when `inputs` is dict or list of dict. Defaults to None.
        output_channel_order (str, optional): The desired image channel order
            of output the data preprocessor. This is also the desired input
            channel order of model (and this most likely to be the output
            order of model). If not passed, no channel order conversion will
            be performed. Defaults to None.
        data_keys (List[str] or str): Keys to preprocess in data samples.
            Defaults to 'gt_img'.
        input_view (tuple, optional): The view of input tensor. This
            argument maybe deleted in the future. Defaults to None.
        output_view (tuple, optional): The view of output tensor. This
            argument maybe deleted in the future. Defaults to None.
        stack_data_sample (bool): Whether stack a list of data samples to one
            data sample. Only support with input data samples are
            `DataSamples`. Defaults to True.
    """
    _NON_IMAGE_KEYS = ['noise']
    _NON_CONCATENATE_KEYS = ['num_batches', 'mode', 'sample_kwargs', 'eq_cfg']

    def __init__(self,
                 mean: Union[Sequence[Union[float, int]], float, int] = 127.5,
                 std: Union[Sequence[Union[float, int]], float, int] = 127.5,
                 pad_size_divisor: int = 1,
                 pad_value: Union[float, int] = 0,
                 pad_mode: str = 'constant',
                 non_image_keys: Optional[Tuple[str, List[str]]] = None,
                 non_concentate_keys: Optional[Tuple[str, List[str]]] = None,
                 output_channel_order: Optional[str] = None,
                 data_keys: Union[List[str], str] = 'gt_img',
                 input_view: Optional[tuple] = None,
                 output_view: Optional[tuple] = None,
                 stack_data_sample=True):

        if not isinstance(mean, (list, tuple)) and mean is not None:
            mean = [mean]
        if not isinstance(std, (list, tuple)) and std is not None:
            std = [std]

        super().__init__(mean, std, pad_size_divisor, pad_value)
        # get channel order
        assert (output_channel_order is None
                or output_channel_order in ['RGB', 'BGR']), (
                    'Only support \'RGB\', \'BGR\' or None for '
                    '\'output_channel_order\', but receive '
                    f'\'{output_channel_order}\'.')
        self.output_channel_order = output_channel_order

        # add user defined keys
        if non_image_keys is not None:
            if not isinstance(non_image_keys, list):
                non_image_keys = [non_image_keys]
            self._NON_IMAGE_KEYS += non_image_keys
        if non_concentate_keys is not None:
            if not isinstance(non_concentate_keys, list):
                non_concentate_keys = [non_concentate_keys]
            self._NON_CONCATENATE_KEYS += non_concentate_keys

        self.pad_mode = pad_mode
        self.pad_size_dict = dict()
        if data_keys is not None and not isinstance(data_keys, list):
            self.data_keys = [data_keys]
        else:
            self.data_keys = data_keys

        # TODO: can be removed since only be used in LIIF
        self.input_view = input_view
        self.output_view = output_view

        self._done_padding = False  # flag for padding checking
        self._conversion_warning_raised = False  # flag for conversion warning

        self.stack_data_sample = stack_data_sample

    def cast_data(self, data: CastData) -> CastData:
        """Copying data to the target device.

        Args:
            data (dict): Data returned by ``DataLoader``.

        Returns:
            CollatedResult: Inputs and data sample at target device.
        """
        if isinstance(data, (str, int, float)):
            return data
        return super().cast_data(data)

    @staticmethod
    def _parse_channel_index(inputs) -> int:
        """Parse channel index of inputs."""
        channel_index_mapping = {2: 1, 3: 0, 4: 1, 5: 2}
        if isinstance(inputs, dict):
            ndim = inputs['fake_img'].ndim
            assert ndim in channel_index_mapping, (
                'Only support (H*W, C), (C, H, W), (N, C, H, W) or '
                '(N, t, C, H, W) inputs. But received '
                f'\'({inputs.shape})\'.')
            channel_index = channel_index_mapping[ndim]
        else:
            assert inputs.ndim in channel_index_mapping, (
                'Only support (H*W, C), (C, H, W), (N, C, H, W) or '
                '(N, t, C, H, W) inputs. But received '
                f'\'({inputs.shape})\'.')
            channel_index = channel_index_mapping[inputs.ndim]

        return channel_index

    def _parse_channel_order(self,
                             key: str,
                             inputs: Tensor,
                             data_sample: Optional[DataSample] = None) -> str:
        channel_index = self._parse_channel_index(inputs)
        if isinstance(inputs, dict):
            num_color_channels = inputs['fake_img'].shape[channel_index]
        else:
            num_color_channels = inputs.shape[channel_index]

        # data sample is None, attempt to infer from input tensor
        if data_sample is None:
            if num_color_channels == 1:
                return 'single'
            else:
                # default as BGR
                return 'BGR'

        # data sample is not None, infer from metainfo
        channel_order_key = 'gt_channel_order' if key == 'gt_img' \
            else f'{key}_channel_order'
        color_type_key = 'gt_color_type' if key == 'gt_img' \
            else f'{key}_color_type'

        # TODO: raise warning here, we can build a dict which fields are keys
        # have been parsed.
        color_flag = data_sample.metainfo.get(color_type_key, None)
        channel_order = data_sample.metainfo.get(channel_order_key, None)

        # handle stacked data sample, refers to `DataSample.stack`
        if isinstance(color_flag, list):
            assert all([c == color_flag[0] for c in color_flag])
            color_flag = color_flag[0]
        if isinstance(channel_order, list):
            assert all([c == channel_order[0] for c in channel_order])
            channel_order = channel_order[0]

        # NOTE: to handle inputs such as Y, users may modify the following code
        if color_flag == 'grayscale':
            assert num_color_channels == 1
            return 'single'
        elif color_flag == 'unchanged':
            # if inputs is not None:
            return 'single' if num_color_channels == 1 else 'BGR'
        else:
            # inference from channel_order
            if channel_order:
                return channel_order
            else:
                # no channel order, infer from num channels
                return 'single' if num_color_channels == 1 else 'BGR'

    def _parse_batch_channel_order(self, key: str, inputs: Sequence,
                                   data_samples: Optional[Sequence[DataSample]]
                                   ) -> str:
        """Parse channel order of inputs in batch."""

        assert len(inputs) == len(data_samples)
        batch_inputs_orders = [
            self._parse_channel_order(key, inp, data_sample)
            for inp, data_sample in zip(inputs, data_samples)
        ]
        inputs_order = batch_inputs_orders[0]

        # security checking for channel order
        assert all([
            inputs_order == order for order in batch_inputs_orders
        ]), (f'Channel order ({batch_inputs_orders}) of input targets '
             f'(\'{key}\') are inconsistent.')

        return inputs_order

    def _update_metainfo(self,
                         padding_info: Tensor,
                         channel_order_info: Optional[dict] = None,
                         data_samples: Optional[SampleList] = None
                         ) -> SampleList:
        """Update `padding_info` and `channel_order` to metainfo of.

        *a batch of `data_samples`*. For channel order, we consider same field
        among data samples share the same channel order. Therefore
        `channel_order` is passed as a dict, which key and value are field
        name and corresponding channel order. For padding info, we consider
        padding info is same among all field of a sample, but can vary between
        samples. Therefore, we pass `padding_info` as Tensor shape like
        (B, 1, 1).

        Args:
            padding_info (Tensor): The padding info of each sample. Shape
                like (B, 1, 1).
            channel_order (dict, Optional): The channel order of target field.
                Key and value are field name and corresponding channel order
                respectively.
            data_samples (List[DataSample], optional): The data samples to
                be updated. If not passed, will initialize a list of empty data
                samples. Defaults to None.

        Returns:
            List[DataSample]: The updated data samples.
        """
        n_samples = padding_info.shape[0]
        if data_samples is None:
            data_samples = [DataSample() for _ in range(n_samples)]
        else:
            assert len(data_samples) == n_samples, (
                f'The length of \'data_samples\'({len(data_samples)}) and '
                f'\'padding_info\'({n_samples}) are inconsistent. Please '
                'check your inputs.')

        # update padding info
        for pad_size, data_sample in zip(padding_info, data_samples):
            data_sample.set_metainfo({'padding_size': pad_size})

        # update channel order
        if channel_order_info is not None:
            for data_sample in data_samples:
                for key, channel_order in channel_order_info.items():
                    data_sample.set_metainfo(
                        {f'{key}_output_channel_order': channel_order})

        self._done_padding = padding_info.sum() != 0
        return data_samples

    def _do_conversion(self,
                       inputs: Tensor,
                       inputs_order: str = 'BGR',
                       target_order: Optional[str] = None
                       ) -> Tuple[Tensor, str]:
        """Conduct channel order conversion for *a batch of inputs*, and return
        the converted inputs and order after conversion.

        inputs_order:
            * RGB / RGB: Convert to target order.
            * SINGLE: Do not change
        """
        if (target_order is None
                or inputs_order.upper() == target_order.upper()):
            # order is not changed, return the input one
            return inputs, inputs_order

        def conversion(inputs, channel_index):
            if inputs.shape[channel_index] == 4:
                new_index = [2, 1, 0, 3]
            else:
                new_index = [2, 1, 0]

            # do conversion
            inputs = torch.index_select(
                inputs, channel_index,
                torch.LongTensor(new_index).to(inputs.device))
            return inputs

        channel_index = self._parse_channel_index(inputs)

        if inputs_order.upper() in ['RGB', 'BGR']:
            inputs = conversion(inputs, channel_index)
            return inputs, target_order
        elif inputs_order.upper() == 'SINGLE':
            if not self._conversion_warning_raised:
                print_log(
                    'Cannot convert inputs with \'single\' channel order '
                    f'to \'output_channel_order\' ({self.output_channel_order}'
                    '). Return without conversion.', 'current', WARNING)
                self._conversion_warning_raised = True
            return inputs, inputs_order
        else:
            raise ValueError(f'Unsupported inputs order \'{inputs_order}\'.')

    def _do_norm(self,
                 inputs: Tensor,
                 do_norm: Optional[bool] = None) -> Tensor:

        do_norm = self._enable_normalize if do_norm is None else do_norm

        if do_norm:
            if self.input_view is None:
                if inputs.ndim == 2:  # special case for (H*W, C) tensor
                    target_shape = [1, -1]
                else:
                    target_shape = [1 for _ in range(inputs.ndim - 3)
                                    ] + [-1, 1, 1]
            else:
                target_shape = self.input_view
            mean = self.mean.view(target_shape)
            std = self.std.view(target_shape)

            # shape checking to avoid broadcast a single channel tensor to 3
            channel_idx = self._parse_channel_index(inputs)
            n_channel_inputs = inputs.shape[channel_idx]
            n_channel_mean = mean.shape[channel_idx]
            n_channel_std = std.shape[channel_idx]
            assert n_channel_mean == 1 or n_channel_mean == n_channel_inputs
            assert n_channel_std == 1 or n_channel_std == n_channel_inputs

            inputs = (inputs - mean) / std
        return inputs

    def _preprocess_image_tensor(self,
                                 inputs: Tensor,
                                 data_samples: Optional[SampleList] = None,
                                 key: str = 'img'
                                 ) -> Tuple[Tensor, SampleList]:
        """Preprocess a batch of image tensor and update metainfo to
        corresponding data samples.

        Args:
            inputs (Tensor): Image tensor with shape (C, H, W), (N, C, H, W) or
                (N, t, C, H, W) to preprocess.
            data_samples (List[DataSample], optional): The data samples
                of corresponding inputs. If not passed, a list of empty data
                samples will be initialized to save metainfo. Defaults to None.
            key (str): The key of image tensor in data samples.
                Defaults to 'img'.

        Returns:
            Tuple[Tensor, List[DataSample]]: The preprocessed image tensor
                and updated data samples.
        """
        if not data_samples:  # none or empty list
            data_samples = [DataSample() for _ in range(inputs.shape[0])]

        assert inputs.dim() in [
            3, 4, 5
        ], ('The input of `_preprocess_image_tensor` should be a (C, H, W), '
            '(N, C, H, W) or (N, t, C, H, W)tensor, but got a tensor with '
            f'shape: {inputs.shape}')
        channel_order = self._parse_batch_channel_order(
            key, inputs, data_samples)
        inputs, output_channel_order = self._do_conversion(
            inputs, channel_order, self.output_channel_order)
        inputs = self._do_norm(inputs)
        h, w = inputs.shape[-2:]
        target_h = math.ceil(h / self.pad_size_divisor) * self.pad_size_divisor
        target_w = math.ceil(w / self.pad_size_divisor) * self.pad_size_divisor
        pad_h = target_h - h
        pad_w = target_w - w
        batch_inputs = F.pad(inputs, (0, pad_w, 0, pad_h), self.pad_mode,
                             self.pad_value)

        padding_size = torch.FloatTensor((0, pad_h, pad_w))[None, ...]
        padding_size = padding_size.repeat(inputs.shape[0], 1)
        data_samples = self._update_metainfo(padding_size,
                                             {key: output_channel_order},
                                             data_samples)
        return batch_inputs, data_samples

    def _preprocess_image_list(self,
                               tensor_list: List[Tensor],
                               data_samples: Optional[SampleList],
                               key: str = 'img') -> Tuple[Tensor, SampleList]:
        """Preprocess a list of image tensor and update metainfo to
        corresponding data samples.

        Args:
            tensor_list (List[Tensor]): Image tensor list to be preprocess.
            data_samples (List[DataSample], optional): The data samples
                of corresponding inputs. If not passed, a list of empty data
                samples will be initialized to save metainfo. Defaults to None.
            key (str): The key of tensor list in data samples.
                Defaults to 'img'.

        Returns:
            Tuple[Tensor, List[DataSample]]: The preprocessed image tensor
                and updated data samples.
        """
        if not data_samples:  # none or empty list
            data_samples = [DataSample() for _ in range(len(tensor_list))]

        channel_order = self._parse_batch_channel_order(
            key, tensor_list, data_samples)
        dim = tensor_list[0].dim()
        assert all([
            tensor.ndim == dim for tensor in tensor_list
        ]), ('Expected the dimensions of all tensors must be the same, '
             f'but got {[tensor.ndim for tensor in tensor_list]}')

        num_img = len(tensor_list)
        all_sizes: torch.Tensor = torch.Tensor(
            [tensor.shape for tensor in tensor_list])
        max_sizes = torch.ceil(
            torch.max(all_sizes, dim=0)[0] /
            self.pad_size_divisor) * self.pad_size_divisor
        padding_sizes = max_sizes - all_sizes
        # The dim of channel and frame index should not be padded.
        padding_sizes[:, :-2] = 0
        if padding_sizes.sum() == 0:
            stacked_tensor = torch.stack(tensor_list)
            stacked_tensor, output_channel_order = self._do_conversion(
                stacked_tensor, channel_order, self.output_channel_order)
            stacked_tensor = self._do_norm(stacked_tensor)
            data_samples = self._update_metainfo(padding_sizes,
                                                 {key: output_channel_order},
                                                 data_samples)
            return stacked_tensor, data_samples

        # `pad` is the second arguments of `F.pad`. If pad is (1, 2, 3, 4),
        # it means that padding the last dim with 1(left) 2(right), padding the
        # penultimate dim to 3(top) 4(bottom). The order of `pad` is opposite
        # of the `padded_sizes`. Therefore, the `padded_sizes` needs to be
        # reversed, and only odd index of pad should be assigned to keep
        # padding "right" and "bottom".
        pad = torch.zeros(num_img, 2 * dim, dtype=torch.int)
        pad[:, 1::2] = padding_sizes[:, range(dim - 1, -1, -1)]
        batch_tensor = []
        for idx, tensor in enumerate(tensor_list):
            paded_tensor = F.pad(tensor, tuple(pad[idx].tolist()),
                                 self.pad_mode, self.pad_value)
            batch_tensor.append(paded_tensor)
        stacked_tensor = torch.stack(batch_tensor)
        stacked_tensor, output_channel_order = self._do_conversion(
            stacked_tensor, channel_order, self.output_channel_order)
        stacked_tensor = self._do_norm(stacked_tensor)
        data_samples = self._update_metainfo(padding_sizes,
                                             {key: output_channel_order},
                                             data_samples)
        # return stacked_tensor, padding_sizes
        return stacked_tensor, data_samples

    def _preprocess_dict_inputs(self,
                                batch_inputs: dict,
                                data_samples: Optional[SampleList] = None
                                ) -> Tuple[dict, SampleList]:
        """Preprocess dict type inputs.

        Args:
            batch_inputs (dict): Input dict.
            data_samples (List[DataSample], optional): The data samples
                of corresponding inputs. If not passed, a list of empty data
                samples will be initialized to save metainfo. Defaults to None.

        Returns:
            Tuple[dict, List[DataSample]]: The preprocessed dict and
                updated data samples.
        """
        pad_size_dict = dict()
        for k, inputs in batch_inputs.items():
            # handle concentrate for values in list
            if isinstance(inputs, list):
                if k in self._NON_CONCATENATE_KEYS:
                    # use the first value
                    assert all([
                        inputs[0] == inp for inp in inputs
                    ]), (f'NON_CONCENTATE_KEY \'{k}\' should be consistency '
                         'among the data list.')
                    batch_inputs[k] = inputs[0]
                else:
                    assert all([
                        isinstance(inp, torch.Tensor) for inp in inputs
                    ]), ('Only support stack list of Tensor in inputs dict. '
                         f'But \'{k}\' is list of \'{type(inputs[0])}\'.')

                    if k not in self._NON_IMAGE_KEYS:
                        # preprocess as image
                        inputs, data_samples = self._preprocess_image_list(
                            inputs, data_samples, k)
                        pad_size_dict[k] = [
                            data.metainfo.get('padding_size')
                            for data in data_samples
                        ]
                    else:
                        # only stack
                        inputs = torch.stack(inputs)

                    batch_inputs[k] = inputs

            elif isinstance(inputs, Tensor) and k not in self._NON_IMAGE_KEYS:
                batch_inputs[k], data_samples = \
                    self._preprocess_image_tensor(inputs, data_samples, k)
                pad_size_dict[k] = [
                    data.metainfo.get('padding_size') for data in data_samples
                ]

        # NOTE: we only support all key shares the same padding size
        if pad_size_dict:
            padding_sizes = list(pad_size_dict.values())[0]
            padding_key = list(pad_size_dict.keys())[0]
            for idx, tar_size in enumerate(padding_sizes):
                for k, sizes in pad_size_dict.items():
                    if (tar_size != sizes[idx]).any():
                        raise ValueError(
                            f'All fields of a data sample should share the '
                            'same padding size, but got different size for '
                            f'\'{k}\'(\'{sizes[idx]}\') and \'{padding_key}\''
                            f'(\'{tar_size}\') at index {idx}.Please check '
                            'your data carefully.')

        return batch_inputs, data_samples

    def _preprocess_data_sample(self, data_samples: SampleList,
                                training: bool) -> DataSample:
        """Preprocess data samples. When `training` is True, fields belong to
        :attr:`self.data_keys` will be converted to
        :attr:`self.output_channel_order` and then normalized by `self.mean`
        and `self.std`. When `training` is False, fields belongs to
        :attr:`self.data_keys` will be attempted to convert to 'BGR' without
        normalization. The corresponding metainfo related to normalization,
        channel order conversion will be updated to data sample as well.

        Args:
            data_samples (List[DataSample]): A list of data samples to
                preprocess.
            training (bool): Whether in training mode.

        Returns:
            list: The list of processed data samples.
        """
        if not training:
            # set default order to BGR in test stage
            target_order, do_norm = 'BGR', False
        else:
            # norm in training, conversion as default (None)
            target_order, do_norm = self.output_channel_order, True

        for data_sample in data_samples:
            if not self.data_keys:
                break
            for key in self.data_keys:
                if not hasattr(data_sample, key):
                    # do not raise error here
                    print_log(f'Cannot find key \'{key}\' in data sample.',
                              'current', WARNING)
                    break

                data = data_sample.get(key)
                data_channel_order = self._parse_channel_order(
                    key, data, data_sample)
                data, channel_order = self._do_conversion(
                    data, data_channel_order, target_order)
                data = self._do_norm(data, do_norm)
                data_sample.set_data({f'{key}': data})
                data_process_meta = {
                    f'{key}_enable_norm': self._enable_normalize,
                    f'{key}_output_channel_order': channel_order,
                    f'{key}_mean': self.mean,
                    f'{key}_std': self.std
                }
                data_sample.set_metainfo(data_process_meta)

        if self.stack_data_sample:
            assert is_seq_of(data_samples, DataSample), (
                'Only support \'stack_data_sample\' for DataSample '
                'object. Please refer to \'DataSample.stack\'.')
            return DataSample.stack(data_samples)
        return data_samples

    def forward(self, data: dict, training: bool = False) -> dict:
        """Performs normalization、padding and channel order conversion.

        Args:
            data (dict): Input data to process.
            training (bool): Whether to in training mode. Default: False.

        Returns:
            dict: Data in the same format as the model input.
        """
        data = self.cast_data(data)
        _batch_inputs = data['inputs']
        _batch_data_samples = data.get('data_samples', None)

        # process input
        if isinstance(_batch_inputs, torch.Tensor):
            _batch_inputs, _batch_data_samples = \
                self._preprocess_image_tensor(
                    _batch_inputs, _batch_data_samples)
        elif is_seq_of(_batch_inputs, torch.Tensor):
            _batch_inputs, _batch_data_samples = \
                self._preprocess_image_list(
                    _batch_inputs, _batch_data_samples)
        elif isinstance(_batch_inputs, dict):
            _batch_inputs, _batch_data_samples = \
                self._preprocess_dict_inputs(
                    _batch_inputs, _batch_data_samples)
        elif is_seq_of(_batch_inputs, dict):
            # convert list of dict to dict of list
            keys = _batch_inputs[0].keys()
            dict_input = {k: [inp[k] for inp in _batch_inputs] for k in keys}
            _batch_inputs, _batch_data_samples = \
                self._preprocess_dict_inputs(
                    dict_input, _batch_data_samples)
        else:
            raise ValueError('Only support following inputs types: '
                             '\'torch.Tensor\', \'List[torch.Tensor]\', '
                             '\'dict\', \'List[dict]\'. But receive '
                             f'\'{type(_batch_inputs)}\'.')
        data['inputs'] = _batch_inputs

        # process data samples
        if _batch_data_samples:
            _batch_data_samples = self._preprocess_data_sample(
                _batch_data_samples, training)

        data['data_samples'] = _batch_data_samples

        return data

    def destruct(self,
                 outputs: Tensor,
                 data_samples: Union[SampleList, DataSample, None] = None,
                 key: str = 'img') -> Union[list, Tensor]:
        """Destruct padding, normalization and convert channel order to BGR if
        could. If `data_samples` is a list, outputs will be destructed as a
        batch of tensor. If `data_samples` is a `DataSample`, `outputs` will be
        destructed as a single tensor.

        Before feed model outputs to visualizer and evaluator, users should
        call this function for model outputs and inputs.

        Use cases:

        >>> # destruct model outputs.
        >>> # model outputs share the same preprocess information with inputs
        >>> # ('img') therefore use 'img' as key
        >>> feats = self.forward_tensor(inputs, data_samples, **kwargs)
        >>> feats = self.data_preprocessor.destruct(feats, data_samples, 'img')

        >>> # destruct model inputs for visualization
        >>> for idx, data_sample in enumerate(data_samples):
        >>>     destructed_input = self.data_preprocessor.destruct(
        >>>         inputs[idx], data_sample, key='img')
        >>>     data_sample.set_data({'input': destructed_input})

        Args:
            outputs (Tensor): Tensor to destruct.
            data_samples (Union[SampleList, DataSample], optional): Data
                samples (or data sample) corresponding to `outputs`.
                Defaults to None
            key (str): The key of field in data sample. Defaults to 'img'.

        Returns:
            Union[list, Tensor]: Destructed outputs.
        """
        # NOTE: only support passing tensor sample, if the output of model is
        # a dict, users should call this manually.
        # Since we do not know whether the outputs is image tensor.
        _batch_outputs = self._destruct_norm_and_conversion(
            outputs, data_samples, key)
        _batch_outputs = self._destruct_padding(_batch_outputs, data_samples)
        _batch_outputs = _batch_outputs.clamp_(0, 255)
        return _batch_outputs

    def _destruct_norm_and_conversion(self, batch_tensor: Tensor,
                                      data_samples: Union[SampleList,
                                                          DataSample, None],
                                      key: str) -> Tensor:
        """De-norm and de-convert channel order. Noted that, we de-norm first,
        and then de-conversion, since mean and std used in normalization is
        based on channel order after conversion.

        Args:
            batch_tensor (Tensor): Tensor to destruct.
            data_samples (Union[SampleList, DataSample], optional): Data
                samples (or data sample) corresponding to `outputs`.
            key (str): The key of field in data sample.

        Returns:
            Tensor: Destructed tensor.
        """

        output_key = f'{key}_output'
        # get channel order from data sample
        if isinstance(data_samples, list):
            inputs_order = self._parse_batch_channel_order(
                output_key, batch_tensor, data_samples)
        else:
            inputs_order = self._parse_channel_order(output_key, batch_tensor,
                                                     data_samples)
        if self._enable_normalize:
            if self.output_view is None:
                if batch_tensor.ndim == 2:  # special case for (H*W, C) tensor
                    target_shape = [1, -1]
                else:
                    target_shape = [1 for _ in range(batch_tensor.ndim - 3)
                                    ] + [-1, 1, 1]
            else:
                target_shape = self.output_view
            mean = self.mean.view(target_shape)
            std = self.std.view(target_shape)
            batch_tensor = batch_tensor * std + mean

        # convert output to 'BGR' if able
        batch_tensor, _ = self._do_conversion(
            batch_tensor, inputs_order=inputs_order, target_order='BGR')

        return batch_tensor

    def _destruct_padding(self,
                          batch_tensor: Tensor,
                          data_samples: Union[SampleList, DataSample, None],
                          same_padding: bool = True) -> Union[list, Tensor]:
        """Destruct padding of the input tensor.

        Args:
            batch_tensor (Tensor): Tensor to destruct.
            data_samples (Union[SampleList, DataSample], optional): Data
                samples (or data sample) corresponding to `outputs`. If
            same_padding (bool): Whether all samples will un-padded with the
                padding info of the first sample, and return a stacked
                un-padded tensor. Otherwise each sample will be unpadded with
                padding info saved in corresponding data samples, and return a
                list of un-padded tensor, since each un-padded tensor may have
                the different shape. Defaults to True.

        Returns:
            Union[list, Tensor]: Destructed outputs.
        """
        # NOTE: If same padding, batch_tensor will un-padded with the padding
        # info # of the first sample and return a Unpadded tensor. Otherwise,
        # input tensor # will un-padded with the corresponding padding info
        # saved in data samples and return a list of tensor.
        if data_samples is None:
            return batch_tensor

        if isinstance(data_samples, list):
            is_batch_data = True
            if 'padding_size' in data_samples[0].metainfo_keys():
                pad_infos = [
                    sample.metainfo['padding_size'] for sample in data_samples
                ]
            else:
                pad_infos = None
        else:
            if 'padding_size' in data_samples.metainfo_keys():
                pad_infos = data_samples.metainfo['padding_size']
            else:
                pad_infos = None
            # NOTE: here we assume padding size in metainfo are saved as tensor
            if not isinstance(pad_infos, list):
                pad_infos = [pad_infos]
                is_batch_data = False
            else:
                is_batch_data = True
            if all([pad_info is None for pad_info in pad_infos]):
                pad_infos = None

        if not is_batch_data:
            batch_tensor = batch_tensor[None, ...]

        if pad_infos is None:
            if self._done_padding:
                print_log(
                    'Cannot find padding information (\'padding_size\') in '
                    'meta info of \'data_samples\'. Please check whether '
                    'you have called \'self.forward\' properly.', 'current',
                    WARNING)
            return batch_tensor if is_batch_data else batch_tensor[0]

        if same_padding:
            # un-pad with the padding info of the first sample
            padded_h, padded_w = pad_infos[0][-2:]
            padded_h, padded_w = int(padded_h), int(padded_w)
            h, w = batch_tensor.shape[-2:]
            batch_tensor = batch_tensor[..., :h - padded_h, :w - padded_w]
            return batch_tensor if is_batch_data else batch_tensor[0]
        else:
            # un-pad with the corresponding padding info
            unpadded_tensors = []
            for idx, pad_info in enumerate(pad_infos):
                padded_h, padded_w = pad_info[-2:]
                padded_h = int(padded_h)
                padded_w = int(padded_w)
                h, w = batch_tensor[idx].shape[-2:]
                unpadded_tensor = batch_tensor[idx][..., :h - padded_h, :w -
                                                    padded_w]
                unpadded_tensors.append(unpadded_tensor)
            return unpadded_tensors if is_batch_data else unpadded_tensors[0]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp
from multiprocessing import Pool

import lmdb
import mmengine
import numpy as np
from PIL import Image
from skimage import img_as_float
from skimage.io import imread, imsave

from mmagic.datasets.transforms import MATLABLikeResize


def export_images(lmdb_path, meta_file_path, out_dir):
    """Export images from lmdb file.

    Ref: https://github.com/fyu/lsun

    Args:
        lmdb_path (str): Lmdb file path.
        meta_file_path (str): Meta file path.
        out_dir (str): Output directory of dataset.
    """

    print('Exporting', lmdb_path, 'to', out_dir)
    env = lmdb.open(
        lmdb_path, map_size=1099511627776, max_readers=100, readonly=True)
    meta_file = open(meta_file_path, 'r')
    img_list = [_.split('.')[0] for _ in meta_file.readlines()]
    txn = env.begin(write=False)
    cursor = txn.cursor()
    count = 0
    for key, val in cursor:
        if key.decode('ascii') in img_list:
            image_out_path = osp.join(out_dir, key.decode('ascii') + '.webp')
            with open(image_out_path, 'wb') as fp:
                fp.write(val)
        count += 1
        if count % 1000 == 0:
            print('Finished', count, 'images')
        if count > len(img_list):
            break
    print('\nFinish exporting images.')


def imresize(img_path, output_path, scale=None, output_shape=None):
    """Resize the image using MATLAB-like downsampling.

    Args:
        img_path (str): Input image path.
        output_path (str): Output image path.
        scale (float | None, optional): The scale factor of the resize
            operation. If None, it will be determined by output_shape.
            Default: None.
        output_shape (tuple(int) | None, optional): The size of the output
            image. If None, it will be determined by scale. Note that if
            scale is provided, output_shape will not be used.
            Default: None.
    """

    matlab_resize = MATLABLikeResize(
        keys=['data'], scale=scale, output_shape=output_shape)
    img = imread(img_path)
    img = img_as_float(img)
    data = {'data': img}
    output = matlab_resize(data)['data']
    output = np.clip(output, 0.0, 1.0) * 255
    output = np.around(output).astype(np.uint8)
    imsave(output_path, output)


def worker(img_name, out_dir):
    """Worker for each process.

    Args:
        img_name (str): Image filename.
        out_dir (str): Output directory of dataset.

    Returns:
        process_info (str): Process information displayed in progress bar.
    """

    _gt_dir = osp.join(out_dir, '_GT')
    gt_dir = osp.join(out_dir, 'GT')
    bix8_dir = osp.join(out_dir, 'BIx8_down')
    bix16_dir = osp.join(out_dir, 'BIx16_down')

    new_img_name = img_name + '.png'
    img = Image.open(osp.join(_gt_dir, img_name + '.webp'))
    img.load()
    img.save(osp.join(_gt_dir, new_img_name))

    imresize(
        osp.join(_gt_dir, new_img_name),
        osp.join(gt_dir, new_img_name),
        output_shape=(256, 256))
    imresize(
        osp.join(gt_dir, new_img_name),
        osp.join(bix8_dir, new_img_name),
        scale=1 / 8)
    imresize(
        osp.join(gt_dir, new_img_name),
        osp.join(bix16_dir, new_img_name),
        scale=1 / 16)

    process_info = f'Processing {new_img_name} ...'
    return process_info


def downsample_images(args):
    """Downsample images for cat_train datasets from the ground-truth."""

    meta_file = open(args.meta_file_path, 'r')
    img_list = [_.split('.')[0] for _ in meta_file.readlines()]
    prog_bar = mmengine.ProgressBar(len(img_list))
    pool = Pool(args.n_thread)
    for path in img_list:
        pool.apply_async(
            worker,
            args=(path, args.out_dir),
            callback=lambda arg: prog_bar.update())
    pool.close()
    pool.join()
    print('All processes done.')


def parse_args():
    parser = argparse.ArgumentParser(
        description='Prepare cat dataset',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--lmdb-path', help='lmdb file path')
    parser.add_argument('--meta-file-path', help='meta file path')
    parser.add_argument('--out-dir', help='output directory of dataset')
    parser.add_argument(
        '--n-thread',
        nargs='?',
        default=8,
        type=int,
        help='thread number when using multiprocessing')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    if not osp.exists(args.out_dir):
        os.makedirs(args.out_dir)
    _gt_dir = osp.join(args.out_dir, '_GT')
    gt_dir = osp.join(args.out_dir, 'GT')
    bix8_dir = osp.join(args.out_dir, 'BIx8_down')
    bix16_dir = osp.join(args.out_dir, 'BIx16_down')
    if not osp.exists(_gt_dir):
        os.makedirs(_gt_dir)
    if not osp.exists(gt_dir):
        os.makedirs(gt_dir)
    if not osp.exists(bix8_dir):
        os.makedirs(bix8_dir)
    if not osp.exists(bix16_dir):
        os.makedirs(bix16_dir)

    # export images
    export_images(args.lmdb_path, args.meta_file_path, _gt_dir)

    # downsample images
    downsample_images(args)



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
from mmcv.transforms import BaseTransform

from mmagic.registry import TRANSFORMS


@TRANSFORMS.register_module()
class MirrorSequence(BaseTransform):
    """Extend short sequences (e.g. Vimeo-90K) by mirroring the sequences.

    Given a sequence with N frames (x1, ..., xN), extend the sequence to
    (x1, ..., xN, xN, ..., x1).

    Required Keys:

    - [KEYS]

    Modified Keys:

    - [KEYS]

    Args:
        keys (list[str]): The frame lists to be extended.
    """

    def __init__(self, keys):

        self.keys = keys

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """

        for key in self.keys:
            if isinstance(results[key], list):
                results[key] = results[key] + results[key][::-1]
            else:
                raise TypeError('The input must be of class list[nparray]. '
                                f'Got {type(results[key])}.')

        return results

    def __repr__(self):

        repr_str = self.__class__.__name__
        repr_str += (f'(keys={self.keys})')

        return repr_str


@TRANSFORMS.register_module()
class TemporalReverse(BaseTransform):
    """Reverse frame lists for temporal augmentation.

    Required keys are the keys in attributes "lq" and "gt",
    added or modified keys are "lq", "gt" and "reverse".

    Args:
        keys (list[str]): The frame lists to be reversed.
        reverse_ratio (float): The probability to reverse the frame lists.
            Default: 0.5.
    """

    def __init__(self, keys, reverse_ratio=0.5):

        self.keys = keys
        self.reverse_ratio = reverse_ratio

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """
        reverse = np.random.random() < self.reverse_ratio

        if reverse:
            for key in self.keys:
                results[key].reverse()

        results['reverse'] = reverse

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(keys={self.keys}, reverse_ratio={self.reverse_ratio})'
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch.nn as nn
from mmengine.model import BaseModule

from mmagic.registry import MODELS
from ..stylegan1 import EqualLinearActModule
from ..stylegan2 import ConvDownLayer, ModMBStddevLayer, ResBlock


@MODELS.register_module()
class MSStyleGAN2Discriminator(BaseModule):
    """StyleGAN2 Discriminator.

    The architecture of this discriminator is proposed in StyleGAN2. More
    details can be found in: Analyzing and Improving the Image Quality of
    StyleGAN CVPR2020.

    Args:
        in_size (int): The input size of images.
        channel_multiplier (int, optional): The multiplier factor for the
            channel number. Defaults to 2.
        blur_kernel (list, optional): The blurry kernel. Defaults
            to [1, 3, 3, 1].
        mbstd_cfg (dict, optional): Configs for minibatch-stddev layer.
            Defaults to dict(group_size=4, channel_groups=1).
    """

    def __init__(self,
                 in_size,
                 channel_multiplier=2,
                 blur_kernel=[1, 3, 3, 1],
                 mbstd_cfg=dict(group_size=4, channel_groups=1),
                 with_adaptive_pool=False,
                 pool_size=(2, 2)):
        super().__init__()
        self.with_adaptive_pool = with_adaptive_pool
        self.pool_size = pool_size

        channels = {
            4: 512,
            8: 512,
            16: 512,
            32: 512,
            64: 256 * channel_multiplier,
            128: 128 * channel_multiplier,
            256: 64 * channel_multiplier,
            512: 32 * channel_multiplier,
            1024: 16 * channel_multiplier,
        }

        log_size = int(np.log2(in_size))
        in_channels = channels[in_size]
        convs = [ConvDownLayer(3, channels[in_size], 1)]

        for i in range(log_size, 2, -1):
            out_channel = channels[2**(i - 1)]
            convs.append(ResBlock(in_channels, out_channel, blur_kernel))

            in_channels = out_channel

        self.convs = nn.Sequential(*convs)
        self.mbstd_layer = ModMBStddevLayer(**mbstd_cfg)

        self.final_conv = ConvDownLayer(in_channels + 1, channels[4], 3)

        if self.with_adaptive_pool:
            self.adaptive_pool = nn.AdaptiveAvgPool2d(pool_size)
            linear_in_channels = channels[4] * pool_size[0] * pool_size[1]
        else:
            linear_in_channels = channels[4] * 4 * 4

        self.final_linear = nn.Sequential(
            EqualLinearActModule(
                linear_in_channels,
                channels[4],
                act_cfg=dict(type='fused_bias')),
            EqualLinearActModule(channels[4], 1),
        )

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor): Input image tensor.

        Returns:
            torch.Tensor: Predict score for the input image.
        """
        x = self.convs(x)

        x = self.mbstd_layer(x)
        x = self.final_conv(x)
        if self.with_adaptive_pool:
            x = self.adaptive_pool(x)
        x = x.view(x.shape[0], -1)
        x = self.final_linear(x)

        return x



# Copyright (c) OpenMMLab. All rights reserved.
from functools import partial
from typing import Dict, List, Optional, Tuple, Union

import mmengine
import numpy as np
import torch
import torch.autograd as autograd
import torch.nn as nn
import torch.nn.functional as F
from mmengine import MessageHub
from mmengine.dist import get_world_size
from mmengine.model import is_model_wrapper
from mmengine.optim import OptimWrapper, OptimWrapperDict
from torch import Tensor

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import ForwardInputs, SampleList
from ...base_models import BaseGAN
from ...utils import get_valid_num_batches, set_requires_grad

ModelType = Union[Dict, nn.Module]
TrainInput = Union[dict, Tensor]


@MODELS.register_module('PGGAN')
@MODELS.register_module()
class ProgressiveGrowingGAN(BaseGAN):
    """Progressive Growing Unconditional GAN.

    In this GAN model, we implement progressive growing training schedule,
    which is proposed in Progressive Growing of GANs for improved Quality,
    Stability and Variation, ICLR 2018.

    We highly recommend to use ``GrowScaleImgDataset`` for saving computational
    load in data pre-processing.

    Notes for **using PGGAN**:

    #. In official implementation, Tero uses gradient penalty with
       ``norm_mode="HWC"``
    #. We do not implement ``minibatch_repeats`` where has been used in
       official Tensorflow implementation.

    Notes for resuming progressive growing GANs:
    Users should specify the ``prev_stage`` in ``train_cfg``. Otherwise, the
    model is possible to reset the optimizer status, which will bring
    inferior performance. For example, if your model is resumed from the
    `256` stage, you should set ``train_cfg=dict(prev_stage=256)``.

    Args:
        generator (dict): Config for generator.
        discriminator (dict): Config for discriminator.
    """

    def __init__(self,
                 generator,
                 discriminator,
                 data_preprocessor,
                 nkimgs_per_scale,
                 noise_size=None,
                 interp_real=None,
                 transition_kimgs: int = 600,
                 prev_stage: int = 0,
                 ema_config: Optional[Dict] = None):
        super().__init__(generator, discriminator, data_preprocessor, 1, 1,
                         noise_size, ema_config)

        # register necessary training status
        self.register_buffer('shown_nkimg', torch.tensor(0.))
        self.register_buffer('_curr_transition_weight', torch.tensor(1.))

        if interp_real is None:
            interp_real = dict(mode='bilinear', align_corners=True)
        self.interp_real_to = partial(F.interpolate, **interp_real)

        self.scales, self.nkimgs = [], []
        for k, v in nkimgs_per_scale.items():
            # support for different data types
            if isinstance(k, str):
                k = (int(k), int(k))
            elif isinstance(k, int):
                k = (k, k)
            else:
                assert mmengine.is_tuple_of(k, int)

            # sanity check for the order of scales
            assert len(self.scales) == 0 or k[0] > self.scales[-1][0]
            self.scales.append(k)
            self.nkimgs.append(v)

        self.cum_nkimgs = np.cumsum(self.nkimgs)
        self.curr_stage = 0
        # dirty workaround for avoiding optimizer bug in resuming
        self.prev_stage = prev_stage
        # actually nkimgs shown at the end of per training stage
        self._actual_nkimgs = []
        # In each scale, transit from previous torgb layer to newer torgb layer
        # with `transition_kimgs` imgs
        self.transition_kimgs = transition_kimgs

        # this buffer is used to resume model easily
        self.register_buffer(
            '_next_scale_int',
            torch.tensor(self.scales[0][0], dtype=torch.int32))
        # TODO: init it with the same value as `_next_scale_int`
        # a dirty workaround for testing
        self.register_buffer(
            '_curr_scale_int',
            torch.tensor(self.scales[-1][0], dtype=torch.int32))

    def forward(self,
                inputs: ForwardInputs,
                data_samples: Optional[list] = None,
                mode: Optional[str] = None) -> SampleList:
        """Sample images from noises by using the generator.

        Args:
            batch_inputs (ForwardInputs): Dict containing the necessary
                information (e.g. noise, num_batches, mode) to generate image.
            data_samples (Optional[list]): Data samples collated by
                :attr:`data_preprocessor`. Defaults to None.
            mode (Optional[str]): `mode` is not used in
                :class:`ProgressiveGrowingGAN`. Defaults to None.

        Returns:
            SampleList: A list of ``DataSample`` contain generated results.
        """
        if isinstance(inputs, Tensor):
            noise = inputs
            curr_scale = transition_weight = None
        else:
            noise = inputs.get('noise', None)
            num_batches = get_valid_num_batches(inputs, data_samples)
            noise = self.noise_fn(noise, num_batches=num_batches)

            curr_scale = inputs.get('curr_scale', None)
            transition_weight = inputs.get('transition_weight', None)
        num_batches = noise.shape[0]

        # use `self.curr_scale` if curr_scale is None
        if curr_scale is None:
            # in training, 'curr_scale' will be set as attribute
            if hasattr(self, 'curr_scale'):
                curr_scale = self.curr_scale[0]
            # in testing, adopt '_curr_scale_int' from buffer as testing scale
            else:
                curr_scale = self._curr_scale_int.item()

        # use `self._curr_transition_weight` if `transition_weight` is None
        if transition_weight is None:
            transition_weight = self._curr_transition_weight.item()

        sample_model = self._get_valid_model(inputs)
        batch_sample_list = []
        if sample_model in ['ema', 'orig']:
            if sample_model == 'ema':
                generator = self.generator_ema
            else:
                generator = self.generator
            outputs = generator(
                noise,
                curr_scale=curr_scale,
                transition_weight=transition_weight)
            outputs = self.data_preprocessor.destruct(outputs, data_samples)

            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples)
            if isinstance(inputs, dict) and 'img' in inputs:
                gen_sample.gt_img = inputs['img']
            gen_sample.fake_img = outputs
            gen_sample.sample_model = sample_model
            gen_sample.noise = noise
            batch_sample_list = gen_sample.split(allow_nonseq_value=True)

        else:  # sample model is 'ema/orig'
            outputs_orig = self.generator(
                noise,
                curr_scale=curr_scale,
                transition_weight=transition_weight)
            outputs_ema = self.generator_ema(
                noise,
                curr_scale=curr_scale,
                transition_weight=transition_weight)
            outputs_orig = self.data_preprocessor.destruct(
                outputs_orig, data_samples)
            outputs_ema = self.data_preprocessor.destruct(
                outputs_ema, data_samples)

            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples)
            if isinstance(inputs, dict) and 'img' in inputs:
                gen_sample.gt_img = inputs['img']
            gen_sample.ema = DataSample(fake_img=outputs_ema)
            gen_sample.orig = DataSample(fake_img=outputs_orig)
            gen_sample.noise = noise
            gen_sample.sample_model = 'ema/orig'
            batch_sample_list = gen_sample.split(allow_nonseq_value=True)

        return batch_sample_list

    def train_discriminator(self, inputs: Tensor,
                            data_samples: List[DataSample],
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Train discriminator.

        Args:
            inputs (Tensor): Inputs from current resolution training.
            data_samples (List[DataSample]): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        real_imgs = inputs
        num_batches = len(data_samples)
        noise_batch = self.noise_fn(num_batches=num_batches)

        with torch.no_grad():
            fake_imgs = self.generator(
                noise_batch,
                curr_scale=self.curr_scale[0],
                transition_weight=self._curr_transition_weight,
                return_noise=False)

        disc_pred_fake = self.discriminator(
            fake_imgs,
            curr_scale=self.curr_scale[0],
            transition_weight=self._curr_transition_weight)
        disc_pred_real = self.discriminator(
            real_imgs,
            curr_scale=self.curr_scale[0],
            transition_weight=self._curr_transition_weight)

        parsed_loss, log_vars = self.disc_loss(disc_pred_fake, disc_pred_real,
                                               fake_imgs, real_imgs)
        optimizer_wrapper.update_params(parsed_loss)
        return log_vars

    def disc_loss(self, disc_pred_fake: Tensor, disc_pred_real: Tensor,
                  fake_data: Tensor, real_data: Tensor) -> Tuple[Tensor, dict]:
        r"""Get disc loss. PGGAN use WGAN-GP's loss and discriminator shift
        loss to train the discriminator.

        .. math:
            L_{D} = \mathbb{E}_{z\sim{p_{z}}}D\left\(G\left\(z\right\)\right\)
                - \mathbb{E}_{x\sim{p_{data}}}D\left\(x\right\) + L_{GP} \\
            L_{GP} = \lambda\mathbb{E}(\Vert\nabla_{\tilde{x}}D(\tilde{x})
                \Vert_2-1)^2 \\
            \tilde{x} = \epsilon x + (1-\epsilon)G(z)
            L_{shift} =

        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.
            disc_pred_real (Tensor): Discriminator's prediction of the real
                images.
            fake_data (Tensor): Generated images, used to calculate gradient
                penalty.
            real_data (Tensor): Real images, used to calculate gradient
                penalty.

        Returns:
            Tuple[Tensor, dict]: Loss value and a dict of log variables.
        """

        losses_dict = dict()
        losses_dict['loss_disc_fake'] = disc_pred_fake.mean()
        losses_dict['loss_disc_real'] = -disc_pred_real.mean()

        # gradient penalty
        batch_size = real_data.size(0)
        alpha = torch.rand(batch_size, 1, 1, 1).to(real_data)

        # interpolate between real_data and fake_data
        interpolates = alpha * real_data + (1. - alpha) * fake_data
        interpolates = autograd.Variable(interpolates, requires_grad=True)

        disc_interpolates = self.discriminator(
            interpolates,
            curr_scale=self.curr_scale[0],
            transition_weight=self._curr_transition_weight)
        gradients = autograd.grad(
            outputs=disc_interpolates,
            inputs=interpolates,
            grad_outputs=torch.ones_like(disc_interpolates),
            create_graph=True,
            retain_graph=True,
            only_inputs=True)[0]
        # norm_mode is 'HWC'
        gradients_penalty = ((
            gradients.reshape(batch_size, -1).norm(2, dim=1) - 1)**2).mean()
        losses_dict['loss_gp'] = 10 * gradients_penalty
        losses_dict['loss_disc_shift'] = 0.001 * 0.5 * (
            disc_pred_fake**2 + disc_pred_real**2)

        parsed_loss, log_vars = self.parse_losses(losses_dict)
        return parsed_loss, log_vars

    def train_generator(self, inputs: Tensor, data_samples: List[DataSample],
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Train generator.

        Args:
            inputs (Tensor): Inputs from current resolution training.
            data_samples (List[DataSample]): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """

        num_batches = len(data_samples)
        noise_batch = self.noise_fn(num_batches=num_batches)

        fake_imgs = self.generator(
            noise_batch,
            num_batches=num_batches,
            curr_scale=self.curr_scale[0],
            transition_weight=self._curr_transition_weight)
        disc_pred_fake_g = self.discriminator(
            fake_imgs,
            curr_scale=self.curr_scale[0],
            transition_weight=self._curr_transition_weight)

        parsed_loss, log_vars = self.gen_loss(disc_pred_fake_g)
        optimizer_wrapper.update_params(parsed_loss)
        return log_vars

    def gen_loss(self, disc_pred_fake: Tensor) -> Tuple[Tensor, dict]:
        r"""Generator loss for PGGAN. PGGAN use WGAN's loss to train the
        generator.

        .. math:
            L_{G} = -\mathbb{E}_{z\sim{p_{z}}}D\left\(G\left\(z\right\)\right\)
                + L_{MSE}

        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.
            recon_imgs (Tensor): Reconstructive images.

        Returns:
            Tuple[Tensor, dict]: Loss value and a dict of log variables.
        """
        losses_dict = dict()
        losses_dict['loss_gen'] = -disc_pred_fake.mean()
        loss, log_vars = self.parse_losses(losses_dict)
        return loss, log_vars

    def train_step(self, data: dict, optim_wrapper: OptimWrapperDict):
        """Train step function.

        This function implements the standard training iteration for
        asynchronous adversarial training. Namely, in each iteration, we first
        update discriminator and then compute loss for generator with the newly
        updated discriminator.

        As for distributed training, we use the ``reducer`` from ddp to
        synchronize the necessary params in current computational graph.

        Args:
            data_batch (dict): Input data from dataloader.
            optimizer (dict): Dict contains optimizer for generator and
                discriminator.
            ddp_reducer (:obj:`Reducer` | None, optional): Reducer from ddp.
                It is used to prepare for ``backward()`` in ddp. Defaults to
                None.
            running_status (dict | None, optional): Contains necessary basic
                information for training, e.g., iteration number. Defaults to
                None.

        Returns:
            dict: Contains 'log_vars', 'num_samples', and 'results'.
        """
        message_hub = MessageHub.get_current_instance()
        curr_iter = message_hub.get_info('iter')

        # update current stage
        self.curr_stage = int(
            min(
                sum(self.cum_nkimgs <= self.shown_nkimg.item()),
                len(self.scales) - 1))
        self.curr_scale = self.scales[self.curr_stage]
        self._curr_scale_int = self._next_scale_int.clone()

        if self.curr_stage != self.prev_stage:
            self.prev_stage = self.curr_stage
            self._actual_nkimgs.append(self.shown_nkimg.item())

        data = self.data_preprocessor(data, True)
        data_sample = data['data_samples']
        real_imgs = data_sample.gt_img

        curr_scale = str(self.curr_scale[0])
        disc_optimizer_wrapper: OptimWrapper = optim_wrapper[
            f'discriminator_{curr_scale}']
        gen_optimizer_wrapper: OptimWrapper = optim_wrapper[
            f'generator_{curr_scale}']
        disc_accu_iters = disc_optimizer_wrapper._accumulative_counts

        # update training configs, like transition weight for torgb layers.
        # get current transition weight for interpolating two torgb layers
        if self.curr_stage == 0:
            transition_weight = 1.
        else:
            transition_weight = (
                self.shown_nkimg.item() -
                self._actual_nkimgs[-1]) / self.transition_kimgs
            # clip to [0, 1]
            transition_weight = min(max(transition_weight, 0.), 1.)
        self._curr_transition_weight = torch.tensor(transition_weight).to(
            self._curr_transition_weight)

        if real_imgs.shape[2:] == self.curr_scale:
            pass
        elif real_imgs.shape[2] >= self.curr_scale[0] and real_imgs.shape[
                3] >= self.curr_scale[1]:
            real_imgs = self.interp_real_to(real_imgs, size=self.curr_scale)
        else:
            raise RuntimeError(
                f'The scale of real image {real_imgs.shape[2:]} is smaller '
                f'than current scale {self.curr_scale}.')

        # normal gan training process
        with disc_optimizer_wrapper.optim_context(self.discriminator):
            log_vars = self.train_discriminator(real_imgs, data_sample,
                                                disc_optimizer_wrapper)

        # add 1 to `curr_iter` because iter is updated in train loop.
        # Whether to update the generator. We update generator with
        # discriminator is fully updated for `self.n_discriminator_steps`
        # iterations. And one full updating for discriminator contains
        # `disc_accu_counts` times of grad accumulations.
        if (curr_iter + 1) % (self.discriminator_steps * disc_accu_iters) == 0:
            set_requires_grad(self.discriminator, False)
            gen_accu_iters = gen_optimizer_wrapper._accumulative_counts

            log_vars_gen_list = []
            # init optimizer wrapper status for generator manually
            gen_optimizer_wrapper.initialize_count_status(
                self.generator, 0, self.generator_steps * gen_accu_iters)
            for _ in range(self.generator_steps * gen_accu_iters):
                with gen_optimizer_wrapper.optim_context(self.generator):
                    log_vars_gen = self.train_generator(
                        real_imgs, data_sample, gen_optimizer_wrapper)

                log_vars_gen_list.append(log_vars_gen)
            log_vars_gen = self.gather_log_vars(log_vars_gen_list)
            log_vars_gen.pop('loss', None)  # remove 'loss' from gen logs

            set_requires_grad(self.discriminator, True)

            # only do ema after generator update
            if self.with_ema_gen and (curr_iter + 1) >= (
                    self.ema_start * self.discriminator_steps *
                    disc_accu_iters):
                self.generator_ema.update_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)
                # if not update buffer, copy buffer from orig model
                if not self.generator_ema.update_buffers:
                    self.generator_ema.sync_buffers(
                        self.generator.module if is_model_wrapper(
                            self.generator) else self.generator)
            elif self.with_ema_gen:
                # before ema, copy weights from orig
                self.generator_ema.sync_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)

            log_vars.update(log_vars_gen)

        # add batch size info to log_vars
        _batch_size = real_imgs.shape[0] * get_world_size()
        self.shown_nkimg += (_batch_size / 1000.)

        log_vars.update(
            dict(
                shown_nkimg=self.shown_nkimg.item(),
                curr_scale=self.curr_scale[0],
                transition_weight=transition_weight))

        # check if a new scale will be added in the next iteration
        _curr_stage = int(
            min(
                sum(self.cum_nkimgs <= self.shown_nkimg.item()),
                len(self.scales) - 1))
        # in the next iteration, we will switch to a new scale
        if _curr_stage != self.curr_stage:
            # `self._next_scale_int` is updated at the end of `train_step`
            self._next_scale_int = self._next_scale_int * 2

        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
import os
from typing import Optional, Sequence, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine import print_log
from torch.utils.data.dataloader import DataLoader
from torchvision import models as torchvision_models

from mmagic.models.utils import get_module_device
from mmagic.registry import METRICS
from ..functional import prepare_vgg_feat
from .base_gen_metric import GenerativeMetric


def compute_pr_distances(row_features,
                         col_features,
                         num_gpus=1,
                         rank=0,
                         col_batch_size=10000):
    r"""Compute distances between real images and fake images.

    This function is used for calculate Precision and Recall metric.
    Refer to:https://github.com/NVlabs/stylegan2-ada-pytorch/blob/main/metrics/precision_recall.py  # noqa
    """
    assert 0 <= rank < num_gpus
    num_cols = col_features.shape[0]
    num_batches = ((num_cols - 1) // col_batch_size // num_gpus + 1) * num_gpus
    col_batches = torch.nn.functional.pad(col_features,
                                          [0, 0, 0, -num_cols % num_batches
                                           ]).chunk(num_batches)
    dist_batches = []
    for col_batch in col_batches[rank::num_gpus]:
        dist_batch = torch.cdist(
            row_features.unsqueeze(0), col_batch.unsqueeze(0))[0]
        for src in range(num_gpus):
            dist_broadcast = dist_batch.clone()
            if num_gpus > 1:
                torch.distributed.broadcast(dist_broadcast, src=src)
            dist_batches.append(dist_broadcast.cpu() if rank == 0 else None)
    return torch.cat(dist_batches, dim=1)[:, :num_cols] if rank == 0 else None


@METRICS.register_module('PR')
@METRICS.register_module()
class PrecisionAndRecall(GenerativeMetric):
    r"""Improved Precision and recall metric.

        In this metric, we draw real and generated samples respectively, and
        embed them into a high-dimensional feature space using a pre-trained
        classifier network. We use these features to estimate the corresponding
        manifold. We obtain the estimation by calculating pairwise Euclidean
        distances between all feature vectors in the set and, for each feature
        vector, construct a hypersphere with radius equal to the distance to its
        kth nearest neighbor. Together, these hyperspheres define a volume in
        the feature space that serves as an estimate of the true manifold.
        Precision is quantified by querying for each generated image whether
        the image is within the estimated manifold of real images.
        Symmetrically, recall is calculated by querying for each real image
        whether the image is within estimated manifold of generated image.

        Ref: https://github.com/NVlabs/stylegan2-ada-pytorch/blob/main/metrics/precision_recall.py  # noqa

        Note that we highly recommend that users should download the vgg16
        script module from the following address. Then, the `vgg16_script` can
        be set with user's local path. If not given, we will use the vgg16 from
        pytorch model zoo. However, this may bring significant different in the
        final results.

        Tero's vgg16: https://nvlabs-fi-cdn.nvidia.com/stylegan2-ada-pytorch/pretrained/metrics/vgg16.pt

        Args:
            num_images (int): The number of evaluated generated samples.
            image_shape (tuple): Image shape in order "CHW". Defaults to None.
            num_real_need (int | None, optional): The number of real images.
                Defaults to None.
            full_dataset (bool, optional): Whether to use full dataset for
                evaluation. Defaults to False.
            k (int, optional): Kth nearest parameter. Defaults to 3.
            bgr2rgb (bool, optional): Whether to change the order of image
                channel. Defaults to True.
            vgg16_script (str, optional): Path for the Tero's vgg16 module.
                Defaults to 'work_dirs/cache/vgg16.pt'.
            row_batch_size (int, optional): The batch size of row data.
                Defaults to 10000.
            col_batch_size (int, optional): The batch size of col data.
                Defaults to 10000.
            auto_save (bool, optional): Whether save vgg feature automatically.
            need_cond_input (bool): If true, the sampler will return the
                conditional input randomly sampled from the original dataset.
                This require the dataset implement `get_data_info` and field
                `gt_label` must be contained in the return value of
                `get_data_info`. Noted that, for unconditional models, set
                `need_cond_input` as True may influence the result of evaluation
                results since the conditional inputs are sampled from the dataset
                distribution; otherwise will be sampled from the uniform
                distribution. Defaults to False.
        """
    name = 'PR'

    def __init__(self,
                 fake_nums,
                 real_nums=-1,
                 k=3,
                 fake_key: Optional[str] = None,
                 real_key: Optional[str] = 'gt_img',
                 need_cond_input: bool = False,
                 sample_model: str = 'ema',
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 vgg16_script='work_dirs/cache/vgg16.pt',
                 vgg16_pkl=None,
                 row_batch_size=10000,
                 col_batch_size=10000,
                 auto_save=True):
        super().__init__(fake_nums, real_nums, fake_key, real_key,
                         need_cond_input, sample_model, collect_device, prefix)
        print_log('loading vgg16 for improved precision and recall...',
                  'current')
        self.vgg16_pkl = vgg16_pkl
        self.vgg16, self.use_tero_scirpt = self._load_vgg(vgg16_script)
        self.k = k

        self.auto_save = auto_save
        self.row_batch_size = row_batch_size
        self.col_batch_size = col_batch_size

    def _load_vgg(self, vgg16_script: Optional[str]) -> Tuple[nn.Module, bool]:
        """Load VGG network from the given path.

        Args:
            vgg16_script: The path of script model of VGG network. If None,
                will load the pytorch version.

        Returns:
            Tuple[nn.Module, str]: The actually loaded VGG network and
                corresponding style.
        """
        if os.path.isfile(vgg16_script):
            vgg16 = torch.jit.load('work_dirs/cache/vgg16.pt').eval()
            use_tero_scirpt = True
        else:
            print_log(
                'Cannot load Tero\'s script module. Use official '
                'vgg16 instead', 'current')
            vgg16 = torchvision_models.vgg16(pretrained=True).eval()
            use_tero_scirpt = False
        return vgg16, use_tero_scirpt

    @torch.no_grad()
    def extract_features(self, images: torch.Tensor) -> torch.Tensor:
        """Extracting image features.

        Args:
            images (torch.Tensor): Images tensor.
        Returns:
            torch.Tensor: Vgg16 features of input images.
        """
        # image must passed in 'bgr'
        images = images[:, [2, 1, 0], ...]
        if self.use_tero_scirpt:
            images = images.to(torch.uint8)
            feature = self.vgg16(images, return_features=True)
        else:
            images = (images - 127.5) / 127.5
            batch = F.interpolate(images, size=(224, 224))
            before_fc = self.vgg16.features(batch)
            before_fc = before_fc.view(-1, 7 * 7 * 512)
            feature = self.vgg16.classifier[:4](before_fc)

        return feature

    @torch.no_grad()
    def compute_metrics(self, results_fake) -> dict:
        """compute_metrics.

        Returns:
            dict: Summarized results.
        """
        gen_features = torch.cat(results_fake, dim=0).to(self.collect_device)
        real_features = self.results_real

        self._result_dict = {}

        for name, manifold, probes in [
            ('precision', real_features, gen_features),
            ('recall', gen_features, real_features)
        ]:
            kth = []
            for manifold_batch in manifold.split(self.row_batch_size):
                distance = compute_pr_distances(
                    row_features=manifold_batch,
                    col_features=manifold,
                    col_batch_size=self.col_batch_size)
                kth.append(
                    distance.to(torch.float32).kthvalue(self.k + 1).values.to(
                        torch.float16))
            kth = torch.cat(kth)
            pred = []
            for probes_batch in probes.split(self.row_batch_size):
                distance = compute_pr_distances(
                    row_features=probes_batch,
                    col_features=manifold,
                    col_batch_size=self.col_batch_size)
                pred.append((distance <= kth).any(dim=1))
            self._result_dict[name] = float(
                torch.cat(pred).to(torch.float32).mean())

        precision = self._result_dict['precision']
        recall = self._result_dict['recall']
        self._result_str = f'precision: {precision}, recall:{recall}'
        return self._result_dict

    @torch.no_grad()
    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.fake_results``, which will be used
        to compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of outputs from the model.
        """
        fake_imgs = []
        for pred in data_samples:
            fake_img_ = pred
            # get ema/orig results
            if self.sample_model in fake_img_:
                fake_img_ = fake_img_[self.sample_model]
            # get specific fake_keys
            if (self.fake_key is not None and self.fake_key in fake_img_):
                fake_img_ = fake_img_[self.fake_key]
            else:
                # get img tensor
                fake_img_ = fake_img_['fake_img']
            fake_imgs.append(fake_img_)
        fake_imgs = torch.stack(fake_imgs, dim=0)
        feat = self.extract_features(fake_imgs)
        feat_list = list(torch.split(feat, 1))
        self.fake_results += feat_list

    @torch.no_grad()
    def prepare(self, module: nn.Module, dataloader: DataLoader) -> None:
        # move to corresponding device
        device = get_module_device(module)
        self.vgg16.to(device)

        vgg_feat = prepare_vgg_feat(dataloader, self, module.data_preprocessor,
                                    self.auto_save)
        if self.real_nums != -1:
            assert self.real_nums <= vgg_feat.shape[0], (
                f'Need \'{self.real_nums}\' of real nums, but only '
                f'\'{vgg_feat.shape[0]}\' images be found in the '
                'inception feature.')
            vgg_feat = vgg_feat[np.random.choice(
                vgg_feat.shape[0], size=self.real_nums, replace=True)]
        self.results_real = vgg_feat



_base_ = ['swinir_x4s48w8d6e180_8xb4-lr2e-4-500k_div2k.py']

experiment_name = 'swinir_x4s64w8d4e60_8xb4-lr2e-4-500k_div2k'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

scale = 4
img_size = 64

# model settings
model = dict(
    generator=dict(
        img_size=img_size,
        depths=[6, 6, 6, 6],
        embed_dim=60,
        num_heads=[6, 6, 6, 6],
        upsampler='pixelshuffledirect'))

# modify patch size of train_dataloader
train_dataloader = _base_.train_dataloader
train_pipeline = train_dataloader['dataset']['pipeline']
train_pipeline[3]['gt_patch_size'] = img_size * scale



_base_ = [
    '../_base_/default_runtime.py',
    '../_base_/datasets/denoising-gaussian_gray_test_config.py'
]

experiment_name = 'swinir_s128w8d6e180_8xb1-lr2e-4-1600k_dfwb-grayDN15'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

# modify sigma of RandomNoise
sigma = 15
test_dataloader = _base_.test_dataloader
for dataloader in test_dataloader:
    test_pipeline = dataloader['dataset']['pipeline']
    test_pipeline[2]['params']['gaussian_sigma'] = [sigma, sigma]

# model settings
model = dict(
    type='BaseEditModel',
    generator=dict(
        type='SwinIRNet',
        upscale=1,
        in_chans=1,
        img_size=128,
        window_size=8,
        img_range=1.0,
        depths=[6, 6, 6, 6, 6, 6],
        embed_dim=180,
        num_heads=[6, 6, 6, 6, 6, 6],
        mlp_ratio=2,
        upsampler='',
        resi_connection='1conv'),
    pixel_loss=dict(type='CharbonnierLoss', eps=1e-9),
    data_preprocessor=dict(type='DataPreprocessor', mean=[0.], std=[255.]))

train_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='grayscale',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='grayscale',
        imdecode_backend='cv2'),
    dict(type='SetValues', dictionary=dict(scale=1)),
    dict(type='PairedRandomCrop', gt_patch_size=128),
    dict(
        type='Flip',
        keys=['img', 'gt'],
        flip_ratio=0.5,
        direction='horizontal'),
    dict(
        type='Flip', keys=['img', 'gt'], flip_ratio=0.5, direction='vertical'),
    dict(type='RandomTransposeHW', keys=['img', 'gt'], transpose_ratio=0.5),
    dict(
        type='RandomNoise',
        params=dict(
            noise_type=['gaussian'],
            noise_prob=[1],
            gaussian_sigma=[sigma, sigma],
            gaussian_gray_noise_prob=0),
        keys=['img']),
    dict(type='PackInputs')
]

val_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='grayscale',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='grayscale',
        imdecode_backend='cv2'),
    dict(
        type='RandomNoise',
        params=dict(
            noise_type=['gaussian'],
            noise_prob=[1],
            gaussian_sigma=[sigma, sigma],
            gaussian_gray_noise_prob=0),
        keys=['img']),
    dict(type='PackInputs')
]

# dataset settings
dataset_type = 'BasicImageDataset'
data_root = 'data'

train_dataloader = dict(
    num_workers=4,
    batch_size=1,
    drop_last=True,
    persistent_workers=False,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type=dataset_type,
        ann_file='meta_info_DFWB8550sub_GT.txt',
        metainfo=dict(dataset_type='dfwb', task_name='denoising'),
        data_root=data_root + '/DFWB',
        data_prefix=dict(img='', gt=''),
        filename_tmpl=dict(img='{}', gt='{}'),
        pipeline=train_pipeline))

val_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type=dataset_type,
        metainfo=dict(dataset_type='set12', task_name='denoising'),
        data_root=data_root + '/Set12',
        data_prefix=dict(img='', gt=''),
        pipeline=val_pipeline))

val_evaluator = [
    dict(type='PSNR', prefix='Set12'),
    dict(type='SSIM', prefix='Set12'),
]

train_cfg = dict(
    type='IterBasedTrainLoop', max_iters=1_600_000, val_interval=5000)
val_cfg = dict(type='ValLoop')

# optimizer
optim_wrapper = dict(
    constructor='DefaultOptimWrapperConstructor',
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=2e-4, betas=(0.9, 0.999)))

# learning policy
param_scheduler = dict(
    type='MultiStepLR',
    by_epoch=False,
    milestones=[800000, 1200000, 1400000, 1500000, 1600000],
    gamma=0.5)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import sys

import mmcv
import torch
from mmengine import Config, print_log
from mmengine.logging import MMLogger
from mmengine.runner import load_checkpoint, set_random_seed

# yapf: disable
sys.path.append(os.path.abspath(os.path.join(__file__, '../..')))  # isort:skip  # noqa

from mmagic.engine import *  # isort:skip  # noqa: F401,F403,E402
from mmagic.datasets import *  # isort:skip  # noqa: F401,F403,E402
from mmagic.models import *  # isort:skip  # noqa: F401,F403,E402

from mmagic.registry import MODELS  # isort:skip  # noqa

# yapf: enable


def parse_args():
    parser = argparse.ArgumentParser(description='Evaluate a GAN model')
    parser.add_argument('config', help='evaluation config file path')
    parser.add_argument('checkpoint', help='checkpoint file')
    parser.add_argument('--seed', type=int, default=2021, help='random seed')
    parser.add_argument(
        '--deterministic',
        action='store_true',
        help='whether to set deterministic options for CUDNN backend.')
    parser.add_argument(
        '--samples-path',
        type=str,
        default='./',
        help='path to store images. If not given, remove it after evaluation\
             finished')
    parser.add_argument(
        '--save-prev-res',
        action='store_true',
        help='whether to store the results from previous stages')
    parser.add_argument(
        '--num-samples',
        type=int,
        default=10,
        help='the number of synthesized samples')
    args = parser.parse_args()
    return args


def _tensor2img(img):
    img = img.permute(1, 2, 0)
    img = img.clamp(0, 255).to(torch.uint8)

    return img.cpu().numpy()


@torch.no_grad()
def main():
    MMLogger.get_instance('mmagic')

    args = parse_args()
    cfg = Config.fromfile(args.config)
    # set cudnn_benchmark
    if cfg.get('cudnn_benchmark', False):
        torch.backends.cudnn.benchmark = True

    # set random seeds
    if args.seed is not None:
        set_random_seed(args.seed, deterministic=args.deterministic)

    # set scope manually
    cfg.model['_scope_'] = 'mmagic'
    # build the model and load checkpoint
    model = MODELS.build(cfg.model)

    model.eval()

    # load ckpt
    print_log(f'Loading ckpt from {args.checkpoint}')
    _ = load_checkpoint(model, args.checkpoint, map_location='cpu')

    # add dp wrapper
    if torch.cuda.is_available():
        model = model.cuda()

    for sample_iter in range(args.num_samples):
        outputs = model.test_step(
            dict(inputs=dict(num_batches=1, get_prev_res=args.save_prev_res)))

        # store results from previous stages
        if args.save_prev_res:
            fake_img = outputs[0].fake_img.data
            prev_res_list = outputs[0].prev_res_list
            prev_res_list.append(fake_img)
            for i, img in enumerate(prev_res_list):
                img = _tensor2img(img)
                mmcv.imwrite(
                    img,
                    os.path.join(args.samples_path, f'stage{i}',
                                 f'rand_sample_{sample_iter}.png'))
        # just store the final result
        else:
            img = _tensor2img(outputs[0].fake_img.data)
            mmcv.imwrite(
                img,
                os.path.join(args.samples_path,
                             f'rand_sample_{sample_iter}.png'))


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import json
import os
import random
from datetime import datetime
from glob import glob

import gradio as gr
import torch
from diffusers import DDIMScheduler, EulerDiscreteScheduler, PNDMScheduler
from mmengine import Config
from safetensors import safe_open

from mmagic.models.editors.animatediff import save_videos_grid
from mmagic.registry import MODELS
from mmagic.utils import register_all_modules

register_all_modules()

sample_idx = 0
scheduler_dict = {
    'Euler': EulerDiscreteScheduler,
    'PNDM': PNDMScheduler,
    'DDIM': DDIMScheduler,
}
cfg = Config.fromfile('configs/animatediff/animatediff_ToonYou.py')
css = """
.toolbutton {
    margin-buttom: 0em 0em 0em 0em;
    max-width: 2.5em;
    min-width: 2.5em !important;
    height: 2.5em;
}
"""


class AnimateController:

    def __init__(self):

        # config dirs
        self.basedir = cfg.models_path
        self.stable_diffusion_dir = os.path.join(self.basedir,
                                                 'StableDiffusion')
        self.motion_module_dir = os.path.join(self.basedir, 'Motion_Module')
        self.personalized_model_dir = os.path.join(self.basedir,
                                                   'DreamBooth_LoRA')
        self.savedir = os.path.join(
            os.getcwd(), 'samples',
            datetime.now().strftime('Gradio-%Y-%m-%dT%H-%M-%S'))
        self.savedir_sample = os.path.join(self.savedir, 'sample')
        os.makedirs(self.savedir, exist_ok=True)

        self.stable_diffusion_list = []
        self.motion_module_list = []
        self.personalized_model_list = []

        self.refresh_stable_diffusion()
        self.refresh_motion_module()
        self.refresh_personalized_model()

        # config models
        self.config = cfg
        self.animatediff = None
        self.lora_model_state_dict = {}

    def refresh_stable_diffusion(self):
        self.stable_diffusion_list = ['runwayml/stable-diffusion-v1-5']

    def refresh_motion_module(self):
        motion_module_list = glob(
            os.path.join(self.motion_module_dir, '*.ckpt'))
        self.motion_module_list = [
            os.path.basename(p) for p in motion_module_list
        ]

    def refresh_personalized_model(self):
        personalized_model_list = glob(
            os.path.join(self.personalized_model_dir, '*.safetensors'))
        self.personalized_model_list = [
            os.path.basename(p) for p in personalized_model_list
        ]

    def update_stable_diffusion(self, stable_diffusion_dropdown):
        self.config['stable_diffusion_v15_url'] = stable_diffusion_dropdown
        self.animatediff = MODELS.build(self.config.model).cuda()
        return gr.Dropdown.update()

    def update_motion_module(self, motion_module_dropdown):
        if self.animatediff.unet is None:
            gr.Info('Please select a pretrained model path.')
            return gr.Dropdown.update(value=None)
        else:
            motion_module_dropdown = os.path.join(self.motion_module_dir,
                                                  motion_module_dropdown)
            self.config.model['motion_module_cfg']['path'] = \
                motion_module_dropdown
            self.animatediff.init_motion_module(
                self.config.model['motion_module_cfg'])
            return gr.Dropdown.update()

    def update_base_model(self, base_model_dropdown):

        if self.animatediff.unet is None:
            gr.Info('Please select a pretrained model path.')
            return gr.Dropdown.update(value=None)
        else:
            base_model_dropdown = os.path.join(self.personalized_model_dir,
                                               base_model_dropdown)
            self.config.model['dream_booth_lora_cfg'][
                'path'] = base_model_dropdown
            self.animatediff.init_dreambooth_lora(
                self.config.model['dream_booth_lora_cfg'])
            return gr.Dropdown.update()

    def update_lora_model(self, lora_model_dropdown):
        lora_model_dropdown = os.path.join(self.personalized_model_dir,
                                           lora_model_dropdown)
        self.lora_model_state_dict = {}
        if lora_model_dropdown == 'none':
            pass
        else:
            with safe_open(
                    lora_model_dropdown, framework='pt', device='cpu') as f:
                for key in f.keys():
                    self.lora_model_state_dict[key] = f.get_tensor(key)
        return gr.Dropdown.update()

    def animate(
            self,
            stable_diffusion_dropdown,
            motion_module_dropdown,
            base_model_dropdown,
            lora_alpha_slider,
            prompt_textbox,
            negative_prompt_textbox,
            # sampler_dropdown,
            sample_step_slider,
            width_slider,
            length_slider,
            height_slider,
            cfg_scale_slider,
            seed_textbox):
        if self.animatediff.unet is None:
            raise gr.Error('Please select a pretrained model path.')
        if motion_module_dropdown == '':
            raise gr.Error('Please select a motion module.')
        if base_model_dropdown == '':
            raise gr.Error('Please select a base DreamBooth model.')
        self.animatediff.cuda()
        self.animatediff.unet.set_use_memory_efficient_attention_xformers(True)

        # TODO: update lora
        # if self.lora_model_state_dict != {}:
        #     pipeline = convert_lora(
        # pipeline,
        # self.lora_model_state_dict,
        # alpha=lora_alpha_slider)

        sample = self.animatediff.infer(
            prompt_textbox,
            negative_prompt=negative_prompt_textbox,
            num_inference_steps=sample_step_slider,
            guidance_scale=cfg_scale_slider,
            width=width_slider,
            height=height_slider,
            video_length=length_slider,
            seed=int(seed_textbox))['samples']

        save_sample_path = os.path.join(self.savedir_sample,
                                        f'{sample_idx}.mp4')
        save_videos_grid(sample, save_sample_path)
        seed = torch.initial_seed()
        sample_config = {
            'prompt': prompt_textbox,
            'n_prompt': negative_prompt_textbox,
            # "sampler": sampler_dropdown, # TODO: More samplers
            'num_inference_steps': sample_step_slider,
            'guidance_scale': cfg_scale_slider,
            'width': width_slider,
            'height': height_slider,
            'video_length': length_slider,
            'seed': seed
        }
        json_str = json.dumps(sample_config, indent=4)
        with open(os.path.join(self.savedir, 'logs.json'), 'a') as f:
            f.write(json_str)
            f.write('\n\n')

        return gr.Video.update(value=save_sample_path)


controller = AnimateController()


def ui():
    with gr.Blocks(css=css) as demo:
        gr.Markdown("""
            # [AnimateDiff: Animate Your Personalized
            # Text-to-Image Diffusion Models without Specific Tuning]
            # (https://arxiv.org/abs/2307.04725)
            Yuwei Guo, Ceyuan Yang*, Anyi Rao, Yaohui Wang,
            Yu Qiao, Dahua Lin, Bo Dai (*Corresponding Author)<br>
            [Arxiv Report](https://arxiv.org/abs/2307.04725) |
            [Project Page](https://animatediff.github.io/) |
            [Github](https://github.com/guoyww/animatediff/)
            """)
        with gr.Column(variant='panel'):
            gr.Markdown("""
                ### 1. Model checkpoints (select pretrained model path first).
                """)
            with gr.Row():
                stable_diffusion_dropdown = gr.Dropdown(
                    label='Pretrained Model Path',
                    choices=controller.stable_diffusion_list,
                    interactive=True,
                )
                stable_diffusion_dropdown.change(
                    fn=controller.update_stable_diffusion,
                    inputs=[stable_diffusion_dropdown],
                    outputs=[stable_diffusion_dropdown])

                stable_diffusion_refresh_button = gr.Button(
                    value='\U0001F503', elem_classes='toolbutton')

                def update_stable_diffusion():
                    controller.refresh_stable_diffusion()
                    return gr.Dropdown.update(
                        choices=controller.stable_diffusion_list)

                stable_diffusion_refresh_button.click(
                    fn=update_stable_diffusion,
                    inputs=[],
                    outputs=[stable_diffusion_dropdown])

            with gr.Row():
                motion_module_dropdown = gr.Dropdown(
                    label='Select motion module',
                    choices=controller.motion_module_list,
                    interactive=True,
                )
                motion_module_dropdown.change(
                    fn=controller.update_motion_module,
                    inputs=[motion_module_dropdown],
                    outputs=[motion_module_dropdown])

                motion_module_refresh_button = gr.Button(
                    value='\U0001F503', elem_classes='toolbutton')

                def update_motion_module():
                    controller.refresh_motion_module()
                    return gr.Dropdown.update(
                        choices=controller.motion_module_list)

                motion_module_refresh_button.click(
                    fn=update_motion_module,
                    inputs=[],
                    outputs=[motion_module_dropdown])

                base_model_dropdown = gr.Dropdown(
                    label='Select base Dreambooth model (required)',
                    choices=controller.personalized_model_list,
                    interactive=True,
                )
                base_model_dropdown.change(
                    fn=controller.update_base_model,
                    inputs=[base_model_dropdown],
                    outputs=[base_model_dropdown])

                lora_model_dropdown = gr.Dropdown(
                    label='Select LoRA model (optional)',
                    choices=['none'] + controller.personalized_model_list,
                    value='none',
                    interactive=True,
                )
                lora_model_dropdown.change(
                    fn=controller.update_lora_model,
                    inputs=[lora_model_dropdown],
                    outputs=[lora_model_dropdown])

                lora_alpha_slider = gr.Slider(
                    label='LoRA alpha',
                    value=0.8,
                    minimum=0,
                    maximum=2,
                    interactive=True)

                personalized_refresh_button = gr.Button(
                    value='\U0001F503', elem_classes='toolbutton')

                def update_personalized_model():
                    controller.refresh_personalized_model()
                    return [
                        gr.Dropdown.update(
                            choices=controller.personalized_model_list),
                        gr.Dropdown.update(choices=['none'] +
                                           controller.personalized_model_list)
                    ]

                personalized_refresh_button.click(
                    fn=update_personalized_model,
                    inputs=[],
                    outputs=[base_model_dropdown, lora_model_dropdown])

        with gr.Column(variant='panel'):
            gr.Markdown("""
                ### 2. Configs for AnimateDiff.
                """)

            prompt_textbox = gr.Textbox(label='Prompt', lines=2)
            negative_prompt_textbox = gr.Textbox(
                label='Negative prompt', lines=2)

            with gr.Row().style(equal_height=False):
                with gr.Column():
                    with gr.Row():
                        sample_step_slider = gr.Slider(
                            label='Sampling steps',
                            value=25,
                            minimum=10,
                            maximum=100,
                            step=1)

                    width_slider = gr.Slider(
                        label='Width',
                        value=512,
                        minimum=256,
                        maximum=1024,
                        step=64)
                    height_slider = gr.Slider(
                        label='Height',
                        value=512,
                        minimum=256,
                        maximum=1024,
                        step=64)
                    length_slider = gr.Slider(
                        label='Animation length',
                        value=16,
                        minimum=8,
                        maximum=24,
                        step=1)
                    cfg_scale_slider = gr.Slider(
                        label='CFG Scale', value=7.5, minimum=0, maximum=20)

                    with gr.Row():
                        seed_textbox = gr.Textbox(label='Seed', value=-1)
                        seed_button = gr.Button(
                            value='\U0001F3B2', elem_classes='toolbutton')
                        seed_button.click(
                            fn=lambda: gr.Textbox.update(
                                value=random.randint(1, 1e8)),
                            inputs=[],
                            outputs=[seed_textbox])

                    generate_button = gr.Button(
                        value='Generate', variant='primary')

                result_video = gr.Video(
                    label='Generated Animation', interactive=False)

            generate_button.click(
                fn=controller.animate,
                inputs=[
                    stable_diffusion_dropdown,
                    motion_module_dropdown,
                    base_model_dropdown,
                    lora_alpha_slider,
                    prompt_textbox,
                    negative_prompt_textbox,
                    # sampler_dropdown,
                    sample_step_slider,
                    width_slider,
                    length_slider,
                    height_slider,
                    cfg_scale_slider,
                    seed_textbox,
                ],
                outputs=[result_video])

    return demo


if __name__ == '__main__':
    demo = ui()
    demo.launch(share=True)



# In this config, we follow the setting `launch_SAGAN_bz128x2_ema.sh` from
# BigGAN's repo. Please refer to https://github.com/ajbrock/BigGAN-PyTorch/blob/master/scripts/launch_SAGAN_bs128x2_ema.sh  # noqa
# In summary, in this config:
# 1. use eps=1e-8 for Spectral Norm
# 2. not use syncBN
# 3. not use  Spectral Norm for embedding layers in cBN
# 4. start EMA at iterations
# 5. use xavier_uniform for weight initialization
# 6. no data augmentation

_base_ = [
    '../_base_/gen_default_runtime.py',
    '../_base_/models/sagan/base_sagan_128x128.py',
    '../_base_/datasets/imagenet_noaug_128.py',
]

# MODEL
init_cfg = dict(type='BigGAN')
model = dict(
    num_classes=1000,
    generator=dict(
        num_classes=1000,
        init_cfg=init_cfg,
        norm_eps=1e-5,
        sn_eps=1e-8,
        auto_sync_bn=False,
        with_embedding_spectral_norm=False),
    discriminator=dict(num_classes=1000, init_cfg=init_cfg, sn_eps=1e-8),
    discriminator_steps=1,
    ema_config=dict(interval=1, momentum=0.999, start_iter=2000))

# TRAINING
train_cfg = dict(
    max_iters=1000000, val_interval=10000, dynamic_intervals=[(800000, 4000)])
train_dataloader = dict(batch_size=32)  # train on 8 gpus

optim_wrapper = dict(
    generator=dict(optimizer=dict(type='Adam', lr=0.0001, betas=(0.0, 0.999))),
    discriminator=dict(
        optimizer=dict(type='Adam', lr=0.0004, betas=(0.0, 0.999))))

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        # vis ema and orig at the same time
        vis_kwargs_list=dict(
            type='Noise',
            name='fake_img',
            sample_model='ema/orig',
            target_keys=['ema.fake_img', 'orig.fake_img']))
]

# METRICS
inception_pkl = './work_dirs/inception_pkl/imagenet-full.pkl'
metrics = [
    dict(
        type='InceptionScore',
        prefix='IS-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='ema'),
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-Full-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        inception_pkl=inception_pkl,
        sample_model='ema')
]
# save multi best checkpoints
default_hooks = dict(
    checkpoint=dict(
        save_best=['FID-Full-50k/fid', 'IS-50k/is'], rule=['less', 'greater']))

val_dataloader = test_dataloader = dict(batch_size=64)
val_evaluator = test_evaluator = dict(metrics=metrics)
